import numpy as np

from .landmark import ASFImporter, PTSImporter


class ImageASFImporter(ASFImporter):
    r"""
    Implements the :meth:`_build_points` method for images. Here, `y` is the
    first axis.

    Parameters
    ----------
    filepath : string
        Absolute filepath of the landmarks
    """

    def __init__(self, filepath):
        super(ImageASFImporter, self).__init__(filepath)

    def _build_points(self, xs, ys):
        """
        For images, `axis 0 = ys` and `axis 1 = xs`. Therefore, return the
        appropriate points array ordering.

        Parameters
        ----------
        xs : (N,) ndarray
            Row vector of `x` coordinates
        ys : (N,) ndarray
            Row vector of `y` coordinates

        Returns
        -------
        points : (N, 2) ndarray
            Array with `ys` as the first axis: `[ys; xs]`
        """
        return np.hstack([ys, xs])


class ImagePTSImporter(PTSImporter):
    r"""
    Implements the :meth:`_build_points` method for images. Here, `y` is the
    first axis.

    Parameters
    ----------
    filepath : string
        Absolute filepath of the landmarks
    """

    def __init__(self, filepath):
        super(ImagePTSImporter, self).__init__(filepath)

    def _build_points(self, xs, ys):
        """
        For images, `axis 0 = ys` and `axis 1 = xs`. Therefore, return the
        appropriate points array ordering.

        Parameters
        ----------
        xs : (N,) ndarray
            Row vector of `x` coordinates
        ys : (N,) ndarray
            Row vector of `y` coordinates

        Returns
        -------
        points : (N, 2) ndarray
            Array with `ys` as the first axis: `[ys; xs]`
        """
        return np.hstack([ys, xs])



import numpy as np
from numpy.testing import assert_allclose, assert_almost_equal
from nose.tools import raises

from menpo.transform import Similarity


def test_basic_2d_similarity():
    linear_component = np.array([[2, -6],
                                 [6, 2]])
    translation_component = np.array([7, -8])
    h_matrix = np.eye(3, 3)
    h_matrix[:-1, :-1] = linear_component
    h_matrix[:-1, -1] = translation_component
    similarity = Similarity(h_matrix)
    x = np.array([[0, 1],
                  [1, 1],
                  [-1, -5],
                  [3, -5]])
    # transform x explicitly
    solution = np.dot(x, linear_component.T) + translation_component
    # transform x using the affine transform
    result = similarity.apply(x)
    # check that both answers are equivalent
    assert_allclose(solution, result)
    # create several copies of x
    x_copies = np.array([x, x, x, x, x, x, x, x])
    # transform all of copies at once using the affine transform
    results = similarity.apply(x_copies)
    # check that all copies have been transformed correctly
    for r in results:
        assert_allclose(solution, r)


def test_similarity_2d_from_vector():
    params = np.array([0.2, 0.1, 1, 2])
    homo = np.array([[params[0] + 1, -params[1], params[2]],
                     [params[1], params[0] + 1, params[3]],
                     [0, 0, 1]])

    sim = Similarity.identity(2).from_vector(params)

    assert_almost_equal(sim.h_matrix, homo)


def test_similarity_2d_as_vector():
    params = np.array([0.2, 0.1, 1.0, 2.0])
    homo = np.array([[params[0] + 1.0, -params[1], params[2]],
                     [params[1], params[0] + 1.0, params[3]],
                     [0.0, 0.0, 1.0]])

    vec = Similarity(homo).as_vector()

    assert_allclose(vec, params)


def test_similarity_2d_n_parameters():
    homo = np.eye(3)
    t = Similarity(homo)
    assert (t.n_parameters == 4)


@raises(NotImplementedError)
def test_similarity_3d_n_parameters_raises_notimplementederror():
    homo = np.eye(4)
    t = Similarity(homo)
    # Raises exception
    t.n_parameters


@raises(NotImplementedError)
def test_similarity_set_h_matrix_raises_notimplementederror():
    s = Similarity(np.eye(3))
    s.set_h_matrix(s.h_matrix)


def test_similarity_identity_2d():
    assert_allclose(Similarity.identity(2).h_matrix,
                    np.eye(3))


def test_similarity_identity_3d():
    assert_allclose(Similarity.identity(3).h_matrix,
                    np.eye(4))



from .convolution import log_gabor
from .decomposition import (eigenvalue_decomposition,
                            principal_component_decomposition)
from .linalg import dot_inplace_left, dot_inplace_right



import numpy as np
from scipy.spatial.distance import cdist
from .base import Transform


class RadialBasisFunction(Transform):
    r"""
    Radial Basis Functions are a class of transform that is used by
    TPS. They have to be able to take their own radial derivative for TPS to
    be able to take it's own total derivative.

    Parameters
    ----------
    c : (n_centres, n_dims) ndarray
        The set of centers that make the basis. Usually represents a set of
        source landmarks.
    """

    def __init__(self, c):
        self.c = c

    @property
    def n_centres(self):
        return self.c.shape[0]

    @property
    def n_dims(self):
        r"""
        The RBF can only be applied on points with the same dimensionality as
        the centres.
        """
        return self.c.shape[1]

    @property
    def n_dims_output(self):
        r"""
        The result of the transform has a dimension (weight) for every centre
        """
        return self.n_centres


class R2LogR2RBF(RadialBasisFunction):
    r"""
    The :math:`r^2 \log{r^2}` basis function.

    The derivative of this function is :math:`2 r (\log{r^2} + 1)`.

    .. note::

        :math:`r = \lVert x - c \rVert`

    Parameters
    ----------
    c : (n_centres, n_dims) ndarray
        The set of centers that make the basis. Usually represents a set of
        source landmarks.
    """

    def __init__(self, c):
        super(R2LogR2RBF, self).__init__(c)

    def _apply(self, x, **kwargs):
        """
        Apply the basis function.

        .. note::

            :math:`r^2 \log{r^2} === r^2 2 \log{r}`

        Parameters
        ----------
        x : (n_points, n_dims) ndarray
            Set of points to apply the basis to.

        Returns
        -------
        u : (n_points, n_centres) ndarray
            The basis function applied to each distance,
            :math:`\lVert x - c \rVert`.
        """
        euclidean_distance = cdist(x, self.c)
        mask = euclidean_distance == 0
        with np.errstate(divide='ignore', invalid='ignore'):
            u = (euclidean_distance ** 2 *
                 (2 * np.log(euclidean_distance)))
        # reset singularities to 0
        u[mask] = 0
        return u


class R2LogRRBF(RadialBasisFunction):
    r"""
    Calculates the :math:`r^2 \log{r}` basis function.

    The derivative of this function is :math:`r (1 + 2 \log{r})`.

    .. note::

        :math:`r = \lVert x - c \rVert`

    Parameters
    ----------
    c : (n_centres, n_dims) ndarray
        The set of centers that make the basis. Usually represents a set of
        source landmarks.
    """

    def __init__(self, c):
        super(R2LogRRBF, self).__init__(c)

    def _apply(self, points, **kwargs):
        """
        Apply the basis function :math:`r^2 \log{r}`.

        Parameters
        ----------
        points : (n_points, n_dims) ndarray
            Set of points to apply the basis to.

        Returns
        -------
        u : (n_points, n_centres) ndarray
            The basis function applied to each distance,
            :math:`\lVert points - c \rVert`.
        """
        euclidean_distance = cdist(points, self.c)
        mask = euclidean_distance == 0
        with np.errstate(divide='ignore', invalid='ignore'):
            u = euclidean_distance ** 2 * np.log(euclidean_distance)
        # reset singularities to 0
        u[mask] = 0
        return u



import numpy as np
from numpy.testing import assert_allclose
import random
import math

from menpo.image import Image, MaskedImage
from menpo.feature import hog, lbp, es, igo, daisy
import menpo.io as mio


def test_imagewindowiterator_hog_padding():
    n_cases = 5
    image_width = np.random.randint(50, 250, [n_cases, 1])
    image_height = np.random.randint(50, 250, [n_cases, 1])
    window_step_horizontal = np.random.randint(1, 10, [n_cases, 1])
    window_step_vertical = np.random.randint(1, 10, [n_cases, 1])
    for i in range(n_cases):
        image = MaskedImage(np.random.randn(image_height[i, 0],
                                            image_width[i, 0], 1))
        hog_im = hog(image,
            mode='dense', window_step_vertical=window_step_vertical[i, 0],
            window_step_horizontal=window_step_horizontal[i, 0],
            window_step_unit='pixels', padding=True)
        n_windows_horizontal = len(range(0, image_width[i, 0],
                                         window_step_horizontal[i, 0]))
        n_windows_vertical = len(range(0, image_height[i, 0],
                                       window_step_vertical[i, 0]))
        assert_allclose(hog_im.shape, (n_windows_vertical,
                                       n_windows_horizontal))


def test_windowiterator_hog_no_padding():
    n_cases = 5
    image_width = np.random.randint(50, 250, [n_cases, 1])
    image_height = np.random.randint(50, 250, [n_cases, 1])
    window_step_horizontal = np.random.randint(1, 10, [n_cases, 1])
    window_step_vertical = np.random.randint(1, 10, [n_cases, 1])
    window_width = np.random.randint(3, 20, [n_cases, 1])
    window_height = np.random.randint(3, 20, [n_cases, 1])
    for i in range(n_cases):
        image = MaskedImage(np.random.randn(image_height[i, 0],
                                            image_width[i, 0], 1))
        hog_img = hog(image,
            mode='dense', cell_size=3, block_size=1,
            window_height=window_height[i, 0], window_width=window_width[i, 0],
            window_unit='pixels',
            window_step_vertical=window_step_vertical[i, 0],
            window_step_horizontal=window_step_horizontal[i, 0],
            window_step_unit='pixels', padding=False)
        n_windows_horizontal = len(range(window_width[i, 0] - 1,
                                         image_width[i, 0],
                                         window_step_horizontal[i, 0]))
        n_windows_vertical = len(range(window_height[i, 0] - 1,
                                       image_height[i, 0],
                                       window_step_vertical[i, 0]))
        assert_allclose(hog_img.shape, (n_windows_vertical,
                                        n_windows_horizontal))


def test_windowiterator_lbp_padding():
    n_cases = 5
    image_width = np.random.randint(50, 250, [n_cases, 1])
    image_height = np.random.randint(50, 250, [n_cases, 1])
    window_step_horizontal = np.random.randint(1, 10, [n_cases, 1])
    window_step_vertical = np.random.randint(1, 10, [n_cases, 1])
    for i in range(n_cases):
        image = MaskedImage(np.random.randn(image_height[i, 0],
                                            image_width[i, 0], 1))
        lbp_img = lbp(image,
            window_step_vertical=window_step_vertical[i, 0],
            window_step_horizontal=window_step_horizontal[i, 0],
            window_step_unit='pixels', padding=True)
        n_windows_horizontal = len(range(0, image_width[i, 0],
                                         window_step_horizontal[i, 0]))
        n_windows_vertical = len(range(0, image_height[i, 0],
                                       window_step_vertical[i, 0]))
        assert_allclose(lbp_img.shape, (n_windows_vertical,
                                          n_windows_horizontal))


def test_windowiterator_lbp_no_padding():
    n_cases = 5
    image_width = np.random.randint(50, 250, [n_cases, 1])
    image_height = np.random.randint(50, 250, [n_cases, 1])
    window_step_horizontal = np.random.randint(1, 10, [n_cases, 1])
    window_step_vertical = np.random.randint(1, 10, [n_cases, 1])
    radius = np.random.randint(3, 5, [n_cases, 1])
    for i in range(n_cases):
        image = Image(np.random.randn(image_height[i, 0],
                                      image_width[i, 0], 1))
        lbp_img = lbp(image,
            radius=radius[i, 0], samples=8,
            window_step_vertical=window_step_vertical[i, 0],
            window_step_horizontal=window_step_horizontal[i, 0],
            window_step_unit='pixels', padding=False)
        window_size = 2 * radius[i, 0] + 1
        n_windows_horizontal = len(range(window_size - 1, image_width[i, 0],
                                         window_step_horizontal[i, 0]))
        n_windows_vertical = len(range(window_size - 1, image_height[i, 0],
                                       window_step_vertical[i, 0]))
        assert_allclose(lbp_img.shape, (n_windows_vertical,
                                        n_windows_horizontal))


def test_hog_channels_dalaltriggs():
    n_cases = 3
    cell_size = np.random.randint(1, 10, [n_cases, 1])
    block_size = np.random.randint(1, 3, [n_cases, 1])
    num_bins = np.random.randint(7, 9, [n_cases, 1])
    channels = np.random.randint(1, 4, [n_cases, 1])
    for i in range(n_cases):
        image = MaskedImage(np.random.randn(40, 40, channels[i, 0]))
        block_size_pixels = cell_size[i, 0] * block_size[i, 0]
        window_width = np.random.randint(block_size_pixels, 40, 1)
        window_height = np.random.randint(block_size_pixels, 40, 1)
        hog_img = hog(image, mode='dense', algorithm='dalaltriggs',
                      cell_size=cell_size[i, 0], block_size=block_size[i, 0],
                      num_bins=num_bins[i, 0], window_height=window_height[0],
                      window_width=window_width[0], window_unit='pixels',
                      window_step_vertical=3, window_step_horizontal=3,
                      window_step_unit='pixels', padding=True)
        length_per_block = block_size[i, 0] * block_size[i, 0] * num_bins[i, 0]
        n_blocks_horizontal = len(range(block_size_pixels - 1, window_width[0],
                                        cell_size[i, 0]))
        n_blocks_vertical = len(range(block_size_pixels - 1, window_height[0],
                                      cell_size[i, 0]))
        n_channels = n_blocks_horizontal * n_blocks_vertical * length_per_block
        assert_allclose(hog_img.n_channels, n_channels)


def test_hog_channels_zhuramanan():
    n_cases = 3
    cell_size = np.random.randint(2, 10, [n_cases, 1])
    channels = np.random.randint(1, 4, [n_cases, 1])
    for i in range(n_cases):
        image = MaskedImage(np.random.randn(40, 40, channels[i, 0]))
        window_width = np.random.randint(3 * cell_size[i, 0], 40, 1)
        window_height = np.random.randint(3 * cell_size[i, 0], 40, 1)
        hog_img = hog(image, mode='dense', algorithm='zhuramanan',
                      cell_size=cell_size[i, 0],
                      window_height=window_height[0],
                      window_width=window_width[0],
                      window_unit='pixels', window_step_vertical=3,
                      window_step_horizontal=3,
                      window_step_unit='pixels', padding=True)
        length_per_block = 31
        n_blocks_horizontal = round(np.float(window_width[0])
                                       / np.float(cell_size[i, 0])) - 2
        n_blocks_vertical = round(np.float(window_height[0])
                                     / np.float(cell_size[i, 0])) - 2
        n_channels = n_blocks_horizontal * n_blocks_vertical * length_per_block
        assert_allclose(hog_img.n_channels, n_channels)


def test_lbp_channels():
    n_cases = 3
    n_combs = np.random.randint(1, 6, [n_cases, 1])
    channels = np.random.randint(1, 4, [n_cases, 1])
    for i in range(n_cases):
        radius = random.sample(xrange(1, 10), n_combs[i, 0])
        samples = random.sample(xrange(4, 12), n_combs[i, 0])
        image = MaskedImage(np.random.randn(40, 40, channels[i, 0]))
        lbp_img = lbp(image, radius=radius, samples=samples,
                      window_step_vertical=3, window_step_horizontal=3,
                      window_step_unit='pixels', padding=True)
        assert_allclose(lbp_img.n_channels, n_combs[i, 0] * channels[i, 0])


def test_igo_channels():
    n_cases = 3
    channels = np.random.randint(1, 10, [n_cases, 1])
    for i in range(n_cases):
        image = Image(np.random.randn(40, 40, channels[i, 0]))
        igo_img = igo(image)
        igo2_img = igo(image, double_angles=True)
        assert_allclose(igo_img.shape, image.shape)
        assert_allclose(igo2_img.shape, image.shape)
        assert_allclose(igo_img.n_channels, 2 * channels[i, 0])
        assert_allclose(igo2_img.n_channels, 4 * channels[i, 0])


def test_es_channels():
    n_cases = 3
    channels = np.random.randint(1, 10, [n_cases, 1])
    for i in range(n_cases):
        image = Image(np.random.randn(40, 40, channels[i, 0]))
        es_img = es(image)
        assert_allclose(es_img.shape, image.shape)
        assert_allclose(es_img.n_channels, 2 * channels[i, 0])


def test_daisy_channels():
    n_cases = 3
    rings = np.random.randint(1, 3, [n_cases, 1])
    orientations = np.random.randint(1, 7, [n_cases, 1])
    histograms = np.random.randint(1, 6, [n_cases, 1])
    channels = np.random.randint(1, 5, [n_cases, 1])
    for i in range(n_cases):
        image = Image(np.random.randn(40, 40, channels[i, 0]))
        daisy_img = daisy(image, step=4, rings=rings[i, 0],
                          orientations=orientations[i, 0],
                          histograms=histograms[i, 0])
        assert_allclose(daisy_img.shape, (3, 3))
        assert_allclose(daisy_img.n_channels,
                        ((rings[i, 0]*histograms[i, 0]+1)*orientations[i, 0]))


def test_igo_values():
    image = Image([[1, 2], [2, 1]])
    igo_img = igo(image)
    res = np.array([
        [[math.cos(math.radians(45)), math.sin(math.radians(45))],
         [math.cos(math.radians(90+45)), math.sin(math.radians(90+45))]],
        [[math.cos(math.radians(-45)), math.sin(math.radians(-45))],
         [math.cos(math.radians(180+45)), math.sin(math.radians(180+45))]]])
    assert_allclose(igo_img.pixels, res)
    image = Image([[0, 0], [0, 0]])
    igo_img = igo(image)
    res = np.array([[[1., 0.], [1., 0.]], [[1., 0.], [1., 0.]]])
    assert_allclose(igo_img.pixels, res)


def test_es_values():
    image = Image([[1, 2], [2, 1]])
    es_img = es(image)
    k = 1 / (2 * (2**0.5))
    res = np.array([[[k, k], [-k, k]], [[k, -k], [-k, -k]]])
    assert_allclose(es_img.pixels, res)
    image = Image([[0, 0], [0, 0]])
    es_img = es(image)
    res = np.array([[[np.nan, np.nan], [np.nan, np.nan]],
                    [[np.nan, np.nan], [np.nan, np.nan]]])
    assert_allclose(es_img.pixels, res)


def test_daisy_values():
    image = Image([[1, 2, 3, 4], [2, 1, 3, 4], [1, 2, 3, 4], [2, 1, 3, 4]])
    daisy_img = daisy(image, step=1, rings=2, radius=1, orientations=8,
                      histograms=8)
    assert_allclose(np.around(daisy_img.pixels[0, 0, 10], 6), 0.000261)
    assert_allclose(np.around(daisy_img.pixels[0, 1, 20], 6), 0.002853)
    assert_allclose(np.around(daisy_img.pixels[1, 0, 30], 6), 0.007266)
    assert_allclose(np.around(daisy_img.pixels[1, 1, 40], 6), 0.001971)


def test_lbp_values():
    image = Image([[0., 6., 0.], [5., 18., 13.], [0., 20., 0.]])
    lbp_img = lbp(image, radius=1, samples=4, mapping_type='none',
                  padding=False)
    assert_allclose(lbp_img.pixels, 8.)
    image = Image([[0., 6., 0.], [5., 25., 13.], [0., 20., 0.]])
    lbp_img = lbp(image, radius=1, samples=4, mapping_type='riu2',
                  padding=False)
    assert_allclose(lbp_img.pixels, 0.)
    image = Image([[0., 6., 0.], [5., 13., 13.], [0., 20., 0.]])
    lbp_img = lbp(image, radius=1, samples=4, mapping_type='u2', padding=False)
    assert_allclose(lbp_img.pixels, 8.)
    image = Image([[0., 6., 0.], [5., 6., 13.], [0., 20., 0.]])
    lbp_img = lbp(image, radius=1, samples=4, mapping_type='ri', padding=False)
    assert_allclose(lbp_img.pixels, 4.)


def test_constrain_landmarks():
    breaking_bad = mio.import_builtin_asset('breakingbad.jpg').as_masked()
    breaking_bad.crop_to_landmarks_inplace(boundary=20)
    breaking_bad = breaking_bad.resize([50, 50])
    breaking_bad.constrain_mask_to_landmarks()
    hog_b = hog(breaking_bad, mode='sparse')
    x = np.where(hog_b.landmarks['PTS'].lms.points[:, 0] > hog_b.shape[1] - 1)
    y = np.where(hog_b.landmarks['PTS'].lms.points[:, 0] > hog_b.shape[0] - 1)
    assert_allclose(len(x[0]) + len(y[0]), 12)
    hog_b = hog(breaking_bad, mode='sparse')
    hog_b.constrain_landmarks_to_bounds()
    x = np.where(hog_b.landmarks['PTS'].lms.points[:, 0] > hog_b.shape[1] - 1)
    y = np.where(hog_b.landmarks['PTS'].lms.points[:, 0] > hog_b.shape[0] - 1)
    assert_allclose(len(x[0]) + len(y[0]), 0)



from .pointcloud import PointCloud
import numpy as np


def mean_pointcloud(pointclouds):
    r"""
    Compute the mean of a `list` of :map:`PointCloud` objects.

    Parameters
    ----------
    pointclouds: `list` of :map:`PointCloud`
        List of point cloud objects from which we want to compute the mean.

    Returns
    -------
    mean_pointcloud : :map:`PointCloud`
        The mean point cloud.
    """
    return PointCloud(np.mean([pc.points for pc in pointclouds], axis=0))



import numpy as np

from menpo.shape import PointCloud
from menpo.transform import Scale

from ..adjacency import mask_adjacency_array, reindex_adjacency_array
from .base import TriMesh


class TexturedTriMesh(TriMesh):
    r"""
    Combines a :map:`TriMesh` with a texture. Also encapsulates the texture
    coordinates required to render the texture on the mesh.

    Parameters
    ----------
    points : ``(n_points, n_dims)`` `ndarray`
        The array representing the points.
    tcoords : ``(N, 2)`` `ndarray`
        The texture coordinates for the mesh.
    texture : :map:`Image`
        The texture for the mesh.
    trilist : ``(M, 3)`` `ndarray` or ``None``, optional
        The triangle list. If ``None``, a Delaunay triangulation of
        the points will be used instead.
    copy: `bool`, optional
        If ``False``, the points, trilist and texture will not be copied on
        assignment.
        In general this should only be used if you know what you are doing.
    """
    def __init__(self, points, tcoords, texture, trilist=None, copy=True):
        super(TexturedTriMesh, self).__init__(points, trilist=trilist,
                                              copy=copy)
        self.tcoords = PointCloud(tcoords, copy=copy)

        if not copy:
            self.texture = texture
        else:
            self.texture = texture.copy()

    def tcoords_pixel_scaled(self):
        r"""
        Returns a :map:`PointCloud` that is modified to be suitable for directly
        indexing into the pixels of the texture (e.g. for manual mapping
        operations). The resulting tcoords behave just like image landmarks
        do.

        The operations that are performed are:

          - Flipping the origin from bottom-left to top-left
          - Scaling the tcoords by the image shape (denormalising them)
          - Permuting the axis so that

        Returns
        -------
        tcoords_scaled : :map:`PointCloud`
            A copy of the tcoords that behave like :map:`Image` landmarks

        Examples
        --------
        Recovering pixel values for every texture coordinate:

        >>> texture = texturedtrimesh.texture
        >>> tc_ps = texturedtrimesh.tcoords_pixel_scaled()
        >>> pixel_values_at_tcs = texture[tc_ps[: ,0], tc_ps[:, 1]]
        """
        scale = Scale(np.array(self.texture.shape)[::-1])
        tcoords = self.tcoords.points.copy()
        # flip the 'y' st 1 -> 0 and 0 -> 1, moving the axis to upper left
        tcoords[:, 1] = 1 - tcoords[:, 1]
        # apply the scale to get the units correct
        tcoords = scale.apply(tcoords)
        # flip axis 0 and axis 1 so indexing is as expected
        tcoords = tcoords[:, ::-1]
        return PointCloud(tcoords)

    def from_vector(self, flattened):
        r"""
        Builds a new :class:`TexturedTriMesh` given the `flattened` 1D vector.
        Note that the trilist, texture, and tcoords will be drawn from self.

        Parameters
        ----------
        flattened : ``(N,)`` `ndarray`
            Vector representing a set of points.

        Returns
        --------
        trimesh : :map:`TriMesh`
            A new trimesh created from the vector with ``self`` trilist.
        """
        return TexturedTriMesh(flattened.reshape([-1, self.n_dims]),
                               self.tcoords.points, self.texture,
                               trilist=self.trilist)

    def from_mask(self, mask):
        """
        A 1D boolean array with the same number of elements as the number of
        points in the TexturedTriMesh. This is then broadcast across the
        dimensions of the mesh and returns a new mesh containing only those
        points that were ``True`` in the mask.

        Parameters
        ----------
        mask : ``(n_points,)`` `ndarray`
            1D array of booleans

        Returns
        -------
        mesh : :map:`TexturedTriMesh`
            A new mesh that has been masked.
        """
        if mask.shape[0] != self.n_points:
            raise ValueError('Mask must be a 1D boolean array of the same '
                             'number of entries as points in this '
                             'TexturedTriMesh.')

        ttm = self.copy()
        if np.all(mask):  # Fast path for all true
            return ttm
        else:
            # Recalculate the mask to remove isolated vertices
            isolated_mask = self._isolated_mask(mask)
            # Recreate the adjacency array with the updated mask
            masked_adj = mask_adjacency_array(isolated_mask, self.trilist)
            ttm.trilist = reindex_adjacency_array(masked_adj)
            ttm.points = ttm.points[isolated_mask, :]
            ttm.tcoords.points = ttm.tcoords.points[isolated_mask, :]
            return ttm

    def _view_3d(self, figure_id=None, new_figure=False, textured=True,
                 **kwargs):
        r"""
        Visualize the :map:`TexturedTriMesh` in 3D.

        Parameters
        ----------
        figure_id : `object`, optional
            The id of the figure to be used.
        new_figure : `bool`, optional
            If ``True``, a new figure is created.
        textured : `bool`, optional
            If `True`, render the texture.

        Returns
        -------
        viewer : :map:`Renderer`
            The viewer object.
        """
        if textured:
            try:
                from menpo3d.visualize import TexturedTriMeshViewer3d
                return TexturedTriMeshViewer3d(
                    figure_id, new_figure, self.points,
                    self.trilist, self.texture,
                    self.tcoords.points).render(**kwargs)
            except ImportError:
                from menpo.visualize import Menpo3dErrorMessage
                raise ImportError(Menpo3dErrorMessage)
        else:
            return super(TexturedTriMesh, self).view(figure_id=figure_id,
                                                     new_figure=new_figure,
                                                     **kwargs)

    def _view_2d(self, figure_id=None, new_figure=False, image_view=True,
                 render_lines=True, line_colour='r', line_style='-',
                 line_width=1., render_markers=True, marker_style='o',
                 marker_size=20, marker_face_colour='k', marker_edge_colour='k',
                 marker_edge_width=1., render_axes=True,
                 axes_font_name='sans-serif', axes_font_size=10,
                 axes_font_style='normal', axes_font_weight='normal',
                 axes_x_limits=None, axes_y_limits=None, figure_size=(10, 8),
                 label=None):
        r"""
        Visualization of the TriMesh in 2D. Currently, explicit textured TriMesh
        viewing is not supported, and therefore viewing falls back to untextured
        2D TriMesh viewing.

        Returns
        -------
        figure_id : `object`, optional
            The id of the figure to be used.
        new_figure : `bool`, optional
            If ``True``, a new figure is created.
        image_view : `bool`, optional
            If ``True`` the TexturedTriMesh will be viewed as if it is in the
            image coordinate system.
        render_lines : `bool`, optional
            If ``True``, the edges will be rendered.
        line_colour : See Below, optional
            The colour of the lines.
            Example options::

                {r, g, b, c, m, k, w}
                or
                (3, ) ndarray

        line_style : ``{-, --, -., :}``, optional
            The style of the lines.
        line_width : `float`, optional
            The width of the lines.
        render_markers : `bool`, optional
            If ``True``, the markers will be rendered.
        marker_style : See Below, optional
            The style of the markers. Example options ::

                {., ,, o, v, ^, <, >, +, x, D, d, s, p, *, h, H, 1, 2, 3, 4, 8}

        marker_size : `int`, optional
            The size of the markers in points^2.
        marker_face_colour : See Below, optional
            The face (filling) colour of the markers.
            Example options ::

                {r, g, b, c, m, k, w}
                or
                (3, ) ndarray

        marker_edge_colour : See Below, optional
            The edge colour of the markers.
            Example options ::

                {r, g, b, c, m, k, w}
                or
                (3, ) ndarray

        marker_edge_width : `float`, optional
            The width of the markers' edge.
        render_axes : `bool`, optional
            If ``True``, the axes will be rendered.
        axes_font_name : See Below, optional
            The font of the axes.
            Example options ::

                {serif, sans-serif, cursive, fantasy, monospace}

        axes_font_size : `int`, optional
            The font size of the axes.
        axes_font_style : {``normal``, ``italic``, ``oblique``}, optional
            The font style of the axes.
        axes_font_weight : See Below, optional
            The font weight of the axes.
            Example options ::

                {ultralight, light, normal, regular, book, medium, roman,
                semibold, demibold, demi, bold, heavy, extra bold, black}

        axes_x_limits : (`float`, `float`) `tuple` or ``None``, optional
            The limits of the x axis.
        axes_y_limits : (`float`, `float`) `tuple` or ``None``, optional
            The limits of the y axis.
        figure_size : (`float`, `float`) `tuple` or ``None``, optional
            The size of the figure in inches.
        label : `str`, optional
            The name entry in case of a legend.

        Returns
        -------
        viewer : :map:`PointGraphViewer2d`
            The viewer object.

        Raises
        ------
        warning
            2D Viewing of Coloured TriMeshes is not supported, automatically
            falls back to 2D :map:`TriMesh` viewing.
        """
        import warnings
        warnings.warn(Warning('2D Viewing of Textured TriMeshes is not '
                              'supported, falling back to TriMesh viewing.'))
        return TriMesh._view_2d(
            self, figure_id=figure_id, new_figure=new_figure,
            image_view=image_view, render_lines=render_lines,
            line_colour=line_colour, line_style=line_style,
            line_width=line_width, render_markers=render_markers,
            marker_style=marker_style, marker_size=marker_size,
            marker_face_colour=marker_face_colour,
            marker_edge_colour=marker_edge_colour,
            marker_edge_width=marker_edge_width, render_axes=render_axes,
            axes_font_name=axes_font_name, axes_font_size=axes_font_size,
            axes_font_style=axes_font_style, axes_font_weight=axes_font_weight,
            axes_x_limits=axes_x_limits, axes_y_limits=axes_y_limits,
            figure_size=figure_size, label=label)

    def __str__(self):
        return '{}\ntexture_shape: {}, n_texture_channels: {}'.format(
            TriMesh.__str__(self), self.texture.shape, self.texture.n_channels)



import abc


class Invertible(object):
    r"""
    Mix-in for invertible transforms. Provides an interface for
    taking the `psuedo` or true inverse of a transform.

    Has to be implemented in conjunction with :map:`Transform`.
    """

    @abc.abstractproperty
    def has_true_inverse(self):
        r"""`True` if the pseudoinverse is an exact inverse.

        :type: `bool`
        """

    def pseudoinverse(self):
        r"""
        The pseudoinverse of the transform - that is, the transform that
        results from swapping source and target, or more formally, negating
        the transforms parameters. If the transform has a true inverse this
        is returned instead.

        :type: ``type(self)``
        """


class VInvertible(Invertible):
    r"""
    Mix-in for :map:`Vectorizable` :map:`Invertible` :map:`Transform` s.

    Prefer this mix-in over :map:`Invertible` if the :map:`Transform` in
    question is :map:`Vectorizable` as this adds :meth:`from_vector` variants
    to the :map:`Invertible` interface. These can be tuned for performance,
    and are, for instance, needed by some of the machinery of fit.
    """
    def pseudoinverse_vector(self, vector):
        r"""
        The vectorized pseudoinverse of a provided vector instance.
        Syntactic sugar for::

            self.from_vector(vector).pseudoinverse().as_vector()

        Can be much faster than the explict call as object creation can be
        entirely avoided in some cases.

        Parameters
        ----------
        vector :  ``(n_parameters,)`` `ndarray`
            A vectorized version of ``self``

        Returns
        -------
        pseudoinverse_vector : ``(n_parameters,)`` `ndarray`
            The pseudoinverse of the vector provided
        """
        return self.from_vector(vector).pseudoinverse().as_vector()



# log_gabor filter and  __frequency_butterworth_filter are derived from Matlab
# scripts written by Peter Kovesi. We maintain his copyright notice below.
#
# Copyright (c) 1999 Peter Kovesi
# School of Computer Science & Software Engineering
# The University of Western Australia
# http://www.csse.uwa.edu.au/
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# The Software is provided "as is", without warranty of any kind.

import numpy as np


def __adjusted_meshgrid(shape):
    """
    Creates an adjusted meshgrid that accounts for odd image sizes. Linearly
    interpolates the values. This meshgrid assumes 'ij' indexing - which is
    due to the 1st dimension of an image being the y-dimension.

    Parameters
    ----------
    shape: tuple
        Size of meshgrid, (M, N, ...). The dimensionality should not be
        swapped due to using images. Therefore, for a 2D image, the expected
        tuple is `(HEIGHT, WIDTH)`.

    Returns
    -------
    meshgrid : list of (M, N, ...) ndarrays
        The meshgrid over each dimension given by the shape.

    """
    adjust_range = []
    for dim in shape:
        adjust_range.append(np.linspace(-0.5, 0.5, dim))

    return np.meshgrid(*adjust_range, indexing='ij')


def __frequency_butterworth_filter(shape, cutoff, order):
    r"""
    Builds an N-D butterworth filter

        ..math::

            f = \frac{1.0}{1.0 + (w / cutoff)^{2n}}

    The frequency origin of the returned filter is at the corners.

    Parameters
    ----------
    shape : tuple
        The size of the filter (M, N, ...)
    cutoff : double
        Cutoff frequency of the filter in the range `[0, 0.5]`
    order : positive int
        Order of the filter. The higher it is the sharper the transition

    Returns
    -------
    butterworth_filter : (M, N, ...) ndarray
        The butterworth filter for the given parameters. Will be the same
        shape as was requested.
    """
    # Dimension-free sum of squares
    grid = __adjusted_meshgrid(shape)
    grid_sq = [g ** 2 for g in grid]
    grid_sq = sum(grid_sq)

    radius = np.sqrt(grid_sq)
    return np.fft.ifftshift(1.0 / ((radius / cutoff) ** (2 * order) + 1.0))


# TODO: merge the 2D and 3D versions if possible
def log_gabor(image, **kwargs):
    r"""
    Creates a log-gabor filter bank, including smoothing the images via a
    low-pass filter at the edges.

    To create a 2D filter bank, simply specify the number of phi
    orientations (orientations in the xy-plane).

    To create a 3D filter bank, you must specify both the number of
    phi (azimuth) and theta (elevation) orientations.

    This algorithm is directly derived from work by Peter Kovesi.

    Parameters
    ----------
    image : ``(M, N, ...)`` `ndarray`
        Image to be convolved
    num_scales : `int`, optional
        Number of wavelet scales.

        ========== ==
        Default 2D 4
        Default 3D 4
        ========== ==
    num_phi_orientations : `int`, optional
        Number of filter orientations in the xy-plane

        ========== ==
        Default 2D 6
        Default 3D 6
        ========== ==
    num_theta_orientations : `int`, optional
        **Only required for 3D**. Number of filter orientations in the z-plane

        ========== ==
        Default 2D N/A
        Default 3D 4
        ========== ==
    min_wavelength : `int`, optional
        Wavelength of smallest scale filter.

        ========== ==
        Default 2D 3
        Default 3D 3
        ========== ==
    scaling_constant : `int`, optional
        Scaling factor between successive filters.

        ========== ==
        Default 2D 2
        Default 3D 2
        ========== ==
    center_sigma : `float`, optional
        Ratio of the standard deviation of the Gaussian describing the Log
        Gabor filter's transfer function in the frequency domain to the filter
        centre frequency.

        ========== ==
        Default 2D 0.65
        Default 3D 0.65
        ========== ==
    d_phi_sigma : `float`, optional
        Angular bandwidth in xy-plane

        ========== ==
        Default 2D 1.3
        Default 3D 1.5
        ========== ==
    d_theta_sigma : `float`, optional
        **Only required for 3D**. Angular bandwidth in z-plane

        ========== ==
        Default 2D N/A
        Default 3D 1.5
        ========== ==

    Returns
    -------
    complex_conv : ``(num_scales, num_orientations, image.shape)`` `ndarray`
        Complex valued convolution results. The real part is the
        result of convolving with the even symmetric filter, the
        imaginary part is the result from convolution with the
        odd symmetric filter.
    bandpass : ``(num_scales, image.shape)`` `ndarray`
        Bandpass images corresponding to each scale `s`
    S : ``(image.shape,)`` `ndarray`
        Convolved image

    Examples
    --------
    Return the magnitude of the convolution over the image at
    scale `s` and orientation `o`

    ::

        np.abs(complex_conv[s, o, :, :])

    Return the phase angles

    ::

        np.angle(complex_conv[s, o, :, :])

    References
    ----------
    .. [1] D. J. Field, "Relations Between the Statistics of Natural Images
        and the Response Properties of Cortical Cells",
        Journal of The Optical Society of America A, Vol 4, No. 12,
        December 1987. pp 2379-2394
    """
    if len(image.shape) == 2:    # 2D filter
        return __log_gabor_2d(image, **kwargs)
    elif len(image.shape) == 3:  # 3D filter
        return __log_gabor_3d(image, **kwargs)
    else:
        raise ValueError("Image must be either 2D or 3D")


def __log_gabor_3d(image, num_scales=4, num_phi_orientations=6,
                   num_theta_orientations=4, min_wavelength=3,
                   scaling_constant=2, center_sigma=0.65, d_theta_sigma=1.5,
                   d_phi_sigma=1.5):
    # Pre-compute sigma values
    theta_sigma = np.pi / num_theta_orientations / d_theta_sigma
    phi_sigma = (2 * np.pi) / num_phi_orientations / d_phi_sigma

    # Allocate space for return structures
    bandpass = np.empty([num_scales, image.shape[0], image.shape[1],
                         image.shape[2]], dtype=np.complex)
    log_gabor = np.empty([num_scales, image.shape[0], image.shape[1],
                          image.shape[2]])
    S = np.zeros(image.shape)
    complex_conv = np.empty([num_scales, num_theta_orientations,
                             num_phi_orientations, image.shape[0],
                             image.shape[1], image.shape[2]], dtype=np.complex)
    tmp_complex_conv = np.empty([num_scales, image.shape[0], image.shape[1],
                                 image.shape[2]], dtype=np.complex)

    # Pre-compute fourier values
    image_fft = np.fft.fftn(image)

    axis0, axis1, axis2 = __adjusted_meshgrid(image.shape)

    radius = np.sqrt(axis0 ** 2 + axis1 ** 2 + axis2 ** 2)
    theta = np.arctan2(axis0, axis1)
    # TODO: Is adding the mean REALLY a good idea?
    m_ab = np.abs(np.mean(radius))
    phi = np.arccos(axis2 / (radius + m_ab))

    radius = np.fft.ifftshift(radius)
    radius[0, 0, 0] = 1.0
    theta = np.fft.ifftshift(theta)
    phi = np.fft.ifftshift(phi)

    sin_theta = np.sin(theta)
    cos_theta = np.cos(theta)
    sin_phi = np.sin(phi)
    cos_phi = np.cos(phi)

    # Compute the lowpass filter
    butterworth_filter = __frequency_butterworth_filter(image.shape, 0.45, 15)

    # Compute radial component of filter
    for s in range(num_scales):
        wavelength = min_wavelength * scaling_constant ** s
        fo = 1.0 / wavelength

        l = np.exp((-np.log(radius / fo) ** 2) /
                   (2.0 * np.log(center_sigma) ** 2))
        l = l * butterworth_filter
        l[0, 0, 0] = 0.0

        log_gabor[s, :, :, :] = l
        bandpass[s, :, :, :] = np.fft.ifft2(image_fft * l)

    # Computer angular component of filter
    for e in range(num_theta_orientations):
        # Pre-compute filter data specific to this orientation
        elevation_angle = e * np.pi / num_theta_orientations

        d_theta_sin = (sin_theta * np.cos(elevation_angle) -
                       cos_theta * np.sin(elevation_angle))
        d_theta_cos = (cos_theta * np.cos(elevation_angle) +
                       sin_theta * np.sin(elevation_angle))
        d_theta = np.abs(np.arctan2(d_theta_sin, d_theta_cos))

        for a in range(num_phi_orientations):
            azimuth_angle = a * 2 * np.pi / num_phi_orientations
            d_phi_sin = (sin_phi * np.cos(azimuth_angle) -
                         cos_phi * np.sin(azimuth_angle))
            d_phi_cos = (cos_phi * np.cos(azimuth_angle) +
                         sin_phi * np.sin(azimuth_angle))
            d_phi = np.abs(np.arctan2(d_phi_sin, d_phi_cos))

            phi_spread = (-d_phi ** 2) / (2 * phi_sigma ** 2)
            theta_spread = (-d_theta ** 2) / (2 * theta_sigma ** 2)
            spread = np.exp(phi_spread + theta_spread)

            # For each scale, multiply by the angular spread
            for s in range(0, num_scales):
                filter_bank = log_gabor[s] * spread

                shifted_filter = np.fft.fftshift(filter_bank)
                S += shifted_filter * np.conjugate(shifted_filter)

                tmp_complex_conv[s, :, :] = np.fft.ifft2(image_fft *
                                                         filter_bank)

            complex_conv[:, e, a, :, :] = tmp_complex_conv[None, None, ...]

    # TODO: Do we need to flip S as in the 2D version?
    return complex_conv, bandpass, S


def __log_gabor_2d(image, num_scales=4, num_orientations=6,
                   min_wavelength=3, scaling_constant=2, center_sigma=0.65,
                   d_phi_sigma=1.3):
    # Allocate space for return structures
    bandpass = np.empty([num_scales, image.shape[0], image.shape[1]],
                        dtype=np.complex)
    log_gabor = np.empty([num_scales, image.shape[0], image.shape[1]])
    S = np.zeros(image.shape)
    complex_conv = np.empty([num_scales, num_orientations, image.shape[0],
                             image.shape[1]], dtype=np.complex)
    tmp_complex_conv = np.empty([num_scales, image.shape[0], image.shape[1]],
                                dtype=np.complex)

    # Pre-compute phi sigma
    phi_sigma = np.pi / num_orientations / d_phi_sigma

    # Pre-compute fourier values
    image_fft = np.fft.fft2(image)

    axis0, axis1 = __adjusted_meshgrid(image.shape)

    radius = np.sqrt(axis0 ** 2 + axis1 ** 2)
    phi = np.arctan2(axis0, axis1)

    radius = np.fft.ifftshift(radius)
    radius[0][0] = 1.0
    phi = np.fft.ifftshift(phi)

    sin_phi = np.sin(phi)
    cos_phi = np.cos(phi)

    # Compute the lowpass filter
    butterworth_filter = __frequency_butterworth_filter(image.shape, 0.45, 15)

    # Compute radial component of filter
    for s in range(num_scales):
        wavelength = min_wavelength * scaling_constant ** s
        fo = 1.0 / wavelength

        l = np.exp((-(np.log(radius / fo)) ** 2) /
                   (2.0 * np.log(center_sigma) ** 2))
        l = l * butterworth_filter
        l[0][0] = 0.0

        log_gabor[s, :, :] = l
        bandpass[s, :, :] = np.fft.ifft2(image_fft * l)

    # Computer angular component of filter
    for o in range(num_orientations):
        # Pre-compute filter data specific to this orientation
        filter_angle = o * np.pi / num_orientations

        ds = (sin_phi * np.cos(filter_angle) -
              cos_phi * np.sin(filter_angle))
        dc = (cos_phi * np.cos(filter_angle) +
              sin_phi * np.sin(filter_angle))

        d_phi = np.abs(np.arctan2(ds, dc))

        # Calculate the standard deviation of the angular Gaussian
        # function used to construct filters in the freq. plane.
        spread = np.exp((-d_phi ** 2.0) / (2.0 * phi_sigma ** 2))

        # For each scale, multiply by the angular spread
        for s in range(0, num_scales):
            filter_bank = log_gabor[s] * spread

            shifted_filter = np.fft.fftshift(filter_bank)
            S += shifted_filter * np.conjugate(shifted_filter)

            tmp_complex_conv[s, :, :] = np.fft.ifft2(image_fft * filter_bank)

        complex_conv[:, o, :, :] = tmp_complex_conv[None, ...]

    # TODO: Why is this done??
    return complex_conv, bandpass, np.flipud(S)



import numpy as np
from mock import patch
from nose.tools import raises
from PIL import Image as PILImage
import menpo.io as mio


@raises(ValueError)
def test_import_incorrect_built_in():
    mio.import_builtin_asset('adskljasdlkajd.obj')


def test_breaking_bad_import():
    img = mio.import_builtin_asset('breakingbad.jpg')
    assert(img.shape == (1080, 1920))
    assert(img.n_channels == 3)
    assert(img.landmarks['PTS'].n_landmarks == 68)


def test_takeo_import():
    img = mio.import_builtin_asset('takeo.ppm')
    assert(img.shape == (225, 150))
    assert(img.n_channels == 3)
    assert(img.landmarks['PTS'].n_landmarks == 68)


def test_einstein_import():
    img = mio.import_builtin_asset('einstein.jpg')
    assert(img.shape == (1024, 817))
    assert(img.n_channels == 1)
    assert(img.landmarks['PTS'].n_landmarks == 68)


def test_lenna_import():
    img = mio.import_builtin_asset('lenna.png')
    assert(img.shape == (512, 512))
    assert(img.n_channels == 3)
    assert(img.landmarks['LJSON'].n_landmarks == 68)


def test_path():
    # choose a random asset (all should have it!)
    img = mio.import_builtin_asset('einstein.jpg')
    path = mio.data_path_to('einstein.jpg')
    assert(img.path == path)
    assert(img.path.stem == 'einstein')
    assert(img.path.suffix == '.jpg')
    assert(img.path.parent == mio.data_dir_path())
    assert(img.path.name == 'einstein.jpg')


def test_import_image():
    img_path = mio.data_dir_path() / 'einstein.jpg'
    im = mio.import_image(img_path)
    assert im.pixels.dtype == np.float
    assert im.n_channels == 1


def test_import_image_no_norm():
    img_path = mio.data_dir_path() / 'einstein.jpg'
    im = mio.import_image(img_path, normalise=False)
    assert im.pixels.dtype == np.uint8


def test_import_landmark_file():
    lm_path = mio.data_dir_path() / 'einstein.pts'
    mio.import_landmark_file(lm_path)


def test_import_images():
    imgs = list(mio.import_images(mio.data_dir_path()))
    imgs_filenames = set(i.path.stem for i in imgs)
    exp_imgs_filenames = {'einstein', 'takeo', 'breakingbad', 'lenna',
                          'menpo_thumbnail'}
    assert(len(exp_imgs_filenames - imgs_filenames) == 0)


def test_ls_builtin_assets():
    assert(set(mio.ls_builtin_assets()) == {'breakingbad.jpg',
                                            'einstein.jpg', 'einstein.pts',
                                            'lenna.png', 'breakingbad.pts',
                                            'lenna.ljson', 'takeo.ppm',
                                            'takeo.pts', 'tongue.jpg',
                                            'tongue.pts',
                                            'menpo_thumbnail.jpg'})


def test_image_paths():
    ls = mio.image_paths(mio.data_dir_path())
    assert(len(list(ls)) == 6)


@raises(ValueError)
def test_import_images_wrong_path_raises_value_error():
    list(mio.import_images('asldfjalkgjlaknglkajlekjaltknlaekstjlakj'))


@raises(ValueError)
def test_import_landmark_files_wrong_path_raises_value_error():
    list(mio.import_landmark_files('asldfjalkgjlaknglkajlekjaltknlaekstjlakj'))


@patch('menpo.io.input.image.PILImage.open')
@patch('menpo.io.input.base.Path.is_file')
def test_importing_RGBA_no_normalise(is_file, mock_image):
    mock_image.return_value = PILImage.new('RGBA', (10, 10))
    is_file.return_value = True

    im = mio.import_image('fake_image_being_mocked.jpg', normalise=False)
    assert im.shape == (10, 10)
    assert im.n_channels == 4
    assert im.pixels.dtype == np.uint8


@patch('menpo.io.input.image.PILImage.open')
@patch('menpo.io.input.base.Path.is_file')
def test_importing_RGBA_normalise(is_file, mock_image):
    from menpo.image import MaskedImage

    mock_image.return_value = PILImage.new('RGBA', (10, 10))
    is_file.return_value = True

    im = mio.import_image('fake_image_being_mocked.jpg', normalise=True)
    assert im.shape == (10, 10)
    assert im.n_channels == 3
    assert im.pixels.dtype == np.float
    assert type(im) == MaskedImage


@patch('menpo.io.input.image.PILImage.open')
@patch('menpo.io.input.base.Path.is_file')
def test_importing_L_no_normalise(is_file, mock_image):
    mock_image.return_value = PILImage.new('L', (10, 10))
    is_file.return_value = True

    im = mio.import_image('fake_image_being_mocked.jpg', normalise=False)
    assert im.shape == (10, 10)
    assert im.n_channels == 1
    assert im.pixels.dtype == np.uint8


@patch('menpo.io.input.image.PILImage.open')
@patch('menpo.io.input.base.Path.is_file')
def test_importing_L_normalise(is_file, mock_image):
    mock_image.return_value = PILImage.new('L', (10, 10))
    is_file.return_value = True

    im = mio.import_image('fake_image_being_mocked.jpg', normalise=True)
    assert im.shape == (10, 10)
    assert im.n_channels == 1
    assert im.pixels.dtype == np.float


@patch('menpo.io.input.image.PILImage.open')
@patch('menpo.io.input.base.Path.is_file')
def test_importing_I_normalise(is_file, mock_image):
    mock_image.return_value = PILImage.new('I', (10, 10))
    is_file.return_value = True

    im = mio.import_image('fake_image_being_mocked.jpg', normalise=True)
    assert im.shape == (10, 10)
    assert im.n_channels == 1
    assert im.pixels.dtype == np.float


@patch('menpo.io.input.image.PILImage.open')
@patch('menpo.io.input.base.Path.is_file')
def test_importing_I_no_normalise(is_file, mock_image):
    mock_image.return_value = PILImage.new('I', (10, 10))
    is_file.return_value = True

    im = mio.import_image('fake_image_being_mocked.jpg', normalise=False)
    assert im.shape == (10, 10)
    assert im.n_channels == 1
    assert im.pixels.dtype == np.int32


@patch('menpo.io.input.image.PILImage.open')
@patch('menpo.io.input.base.Path.is_file')
def test_importing_1_normalise(is_file, mock_image):
    from menpo.image import BooleanImage

    mock_image.return_value = PILImage.new('1', (10, 10))
    is_file.return_value = True

    im = mio.import_image('fake_image_being_mocked.jpg', normalise=True)
    assert im.shape == (10, 10)
    assert im.n_channels == 1
    assert im.pixels.dtype == np.bool
    assert type(im) == BooleanImage


@patch('menpo.io.input.image.PILImage.open')
@patch('menpo.io.input.base.Path.is_file')
def test_importing_1_no_normalise(is_file, mock_image):
    from menpo.image import BooleanImage

    mock_image.return_value = PILImage.new('1', (10, 10))
    is_file.return_value = True

    im = mio.import_image('fake_image_being_mocked.jpg', normalise=False)
    assert im.shape == (10, 10)
    assert im.n_channels == 1
    assert im.pixels.dtype == np.bool
    assert type(im) == BooleanImage


@patch('menpo.io.input.image.PILImage.open')
@patch('menpo.io.input.base.Path.is_file')
def test_importing_P_normalise(is_file, mock_image):
    mock_image.return_value = PILImage.new('P', (10, 10))
    is_file.return_value = True

    im = mio.import_image('fake_image_being_mocked.jpg', normalise=True)
    assert im.shape == (10, 10)
    assert im.n_channels == 3
    assert im.pixels.dtype == np.float


@patch('menpo.io.input.image.PILImage.open')
@patch('menpo.io.input.base.Path.is_file')
def test_importing_P_no_normalise(is_file, mock_image):
    mock_image.return_value = PILImage.new('P', (10, 10))
    is_file.return_value = True

    im = mio.import_image('fake_image_being_mocked.jpg', normalise=False)
    assert im.shape == (10, 10)
    assert im.n_channels == 3
    assert im.pixels.dtype == np.uint8


@patch('menpo.io.input.image.PILImage.open')
@patch('menpo.io.input.base.Path.is_file')
def test_importing_GIF_normalise(is_file, mock_image):
    mock_image.return_value = PILImage.new('P', (10, 10))
    is_file.return_value = True

    im = mio.import_image('fake_image_being_mocked.gif', normalise=True)
    assert im.shape == (10, 10)
    assert im.n_channels == 3
    assert im.pixels.dtype == np.float


@patch('menpo.io.input.image.PILImage.open')
@patch('menpo.io.input.base.Path.is_file')
def test_importing_GIF_no_normalise(is_file, mock_image):
    mock_image.return_value = PILImage.new('P', (10, 10))
    is_file.return_value = True

    im = mio.import_image('fake_image_being_mocked.gif', normalise=False)
    assert im.shape == (10, 10)
    assert im.n_channels == 3
    assert im.pixels.dtype == np.uint8


@patch('menpo.io.input.image.PILImage.open')
@patch('menpo.io.input.base.Path.is_file')
@raises(ValueError)
def test_importing_GIF_non_pallete_exception(is_file, mock_image):
    mock_image.return_value = PILImage.new('RGB', (10, 10))
    is_file.return_value = True

    mio.import_image('fake_image_being_mocked.gif', normalise=False)



import abc
import numpy as np
from copy import deepcopy
from menpo.base import Copyable
from menpo.transform.base import Alignment, Invertible, Transform
from .fastpwa import CLookupPWA
# TODO View is broken for PWA (TriangleContainmentError)


class TriangleContainmentError(Exception):
    r"""
    Exception that is thrown when an attempt is made to map a point with a
    PWATransform that does not lie in a source triangle.

    points_outside_source_domain : (d,) ndarray
        A boolean value for the d points that were attempted to be applied.
        If True, the point was outside of the domain.
    """
    def __init__(self, points_outside_source_domain):
        super(TriangleContainmentError, self).__init__()
        self.points_outside_source_domain = points_outside_source_domain


def containment_from_alpha_beta(alpha, beta):
    r"""
    Check `alpha` and `beta` are within a triangle (`alpha >= 0`,
    `beta >= 0`, `alpha + beta <= 1`). Returns the indices of the
    triangles that are `alpha` and `beta` are in. If any of the
    points are not contained in a triangle,
    raises a TriangleContainmentError.

    Parameters
    ----------
    alpha: (K, `n_tris`) ndarray
        Alpha for each point and triangle being tested.
    beta: (K, `n_tris`) ndarray
        Beta for each point and triangle being tested.

    Returns
    -------
    tri_index : (L,) ndarray
        triangle index for each `points`, assigning each
        point in a triangle to the triangle index.

    Raises
    ------
    TriangleContainmentError
    All `points` must be contained in a source triangle. Check
    `error.points_outside_source_domain` to handle this case.
    """
    # (K, n_tris), boolean for whether a given triangle contains a given
    #  point
    point_containment = np.logical_and(np.logical_and(
        alpha >= 0, beta >= 0),
        alpha + beta <= 1)
    # is each point in a triangle?
    point_in_a_triangle = np.any(point_containment, axis=1)
    if np.any(~point_in_a_triangle):
        raise TriangleContainmentError(~point_in_a_triangle)
    point_index, tri_index = np.nonzero(point_containment)
    # don't want duplicates! ensure that here:
    index = np.zeros(alpha.shape[0])
    index[point_index] = tri_index
    return index.astype(np.uint32)


def alpha_beta(i, ij, ik, points):
    r"""
    Calculates the alpha and beta values (barycentric coordinates) for each
    triangle for all points provided. Note that this does not raise a
    TriangleContainmentError.

    Parameters
    ----------
    i : (`n_tris`, 2) ndarray
        The coordinate of the i'th point of each triangle

    ij (`n_tris`, 2) ndarray
        The vector between the i'th point and the j'th point of each
        triangle

    ik (`n_tris`, 2) ndarray
        The vector between the i'th point and the k'th point of each
        triangle

    points : (`n_points`, 2) ndarray
        Points to calculate the barycentric coordinates for.

    Returns
    --------
    alpha : (`n_points`, `n_tris`)
        The alpha for each point and triangle. Alpha can be interpreted
        as the contribution of the ij vector to the position of the
        point in question.
    beta : (`n_points`, `n_tris`)
        The beta for each point and triangle. Beta can be interpreted as
         the contribution of the ik vector to the position of the point
         in question.
    """
    ip = points[..., None] - i
    dot_jj = np.einsum('dt, dt -> t', ij, ij)
    dot_kk = np.einsum('dt, dt -> t', ik, ik)
    dot_jk = np.einsum('dt, dt -> t', ij, ik)
    dot_pj = np.einsum('vdt, dt -> vt', ip, ij)
    dot_pk = np.einsum('vdt, dt -> vt', ip, ik)

    d = 1.0/(dot_jj * dot_kk - dot_jk * dot_jk)
    alpha = (dot_kk * dot_pj - dot_jk * dot_pk) * d
    beta = (dot_jj * dot_pk - dot_jk * dot_pj) * d
    return alpha, beta


def index_alpha_beta(i, ij, ik, points):
    """
    Finds for each input point the index of it's bounding triangle
    and the alpha and beta value for that point in the triangle. Note
    this means that the following statements will always be true:
        alpha + beta <= 1
        alpha >= 0
        beta >= 0
    for each triangle result.
    Trying to map a point that does not exist in a
    triangle throws a TriangleContainmentError.

    Parameters
    ----------
    i : (`n_tris`, 2) ndarray
        The coordinate of the i'th point of each triangle

    ij (`n_tris`, 2) ndarray
        The vector between the i'th point and the j'th point of each
        triangle

    ik (`n_tris`, 2) ndarray
        The vector between the i'th point and the k'th point of each
        triangle

    points : (`n_points`, 2) ndarray
        Points to calculate the barycentric coordinates for.


    Returns
    -------
    tri_index : (`n_tris`,) ndarray
        triangle index for each of the `points`, assigning each
        point to it's containing triangle.
    alpha : (`n_tris`,) ndarray
        Alpha for containing triangle of each point.
    beta : (`n_tris`,) ndarray
        Beta for containing triangle of each point.

    Raises
    ------
    TriangleContainmentError
    All `points` must be contained in a source triangle. Check
    `error.points_outside_source_domain` to handle this case.
    """
    alpha, beta = alpha_beta(i, ij, ik, points)
    each_point = np.arange(points.shape[0])
    index = containment_from_alpha_beta(alpha, beta)
    return index, alpha[each_point, index], beta[each_point, index]


def barycentric_vectors(points, trilist):
    r"""
    Compute the affine transformation between each triangle in the source
    and target. This is calculated analytically.

    Parameters
    ----------

    points : (`n_points`, 2) ndarray
        Points to calculate the barycentric coordinates for.

    trilist: (`n_tris`, 3) ndarray
        The 0-based index triangulation joining the points.

    Returns
    -------
    i : (`n_tris`, 2) ndarray
        The coordinate of the i'th point of each triangle

    ij (`n_tris`, 2) ndarray
        The vector between the i'th point and the j'th point of each
        triangle

    ik (`n_tris`, 2) ndarray
        The vector between the i'th point and the k'th point of each
        triangle
    """
    # we permute the axes of the indexed point set to have shape
    # [3, n_dims, n_tris] for ease of indexing in.
    x = np.transpose(points[trilist],  axes=[1, 2, 0])
    return x[0], x[1] - x[0], x[2] - x[0]


# Note we inherit from Alignment first to get it's n_dims behavior
class AbstractPWA(Alignment, Transform, Invertible):
    r"""
    A piecewise affine transformation. This is composed of a number of
    triangles defined be a set of source and target vertices. These vertices
    are related by a common triangle list. No limitations on the nature of
    the triangle list are imposed. Points can then be mapped via
    barycentric coordinates from the source to the target space.
    Trying to map points that are not contained by any source triangle
    throws a TriangleContainmentError, which contains diagnostic information.

    Parameters
    ----------
    source : :class:`menpo.shape.PointCloud` or :class:`menpo.shape.TriMesh`
        The source points. If a TriMesh is provided, the triangulation on
        the TriMesh is used. If a :class:`menpo.shape.PointCloud`
        is provided, a Delaunay triangulation of the source is performed
        automatically.
    target : :class:`PointCloud`
        The target points. Note that the trilist is entirely decided by
        the source.

    Raises
    ------
    ValueError
        Source and target must both be 2D.

    TriangleContainmentError
        All points to apply must be contained in a source triangle. Check
        `error.points_outside_source_domain` to handle this case.
    """
    def __init__(self, source, target):
        from menpo.shape import TriMesh  # to avoid circular import
        if not isinstance(source, TriMesh):
            source = TriMesh(source.points)
        Alignment.__init__(self, source, target)
        if self.n_dims != 2:
            raise ValueError("source and target must be 2 "
                             "dimensional")
        self.ti, self.tij, self.tik = None, None, None
        self._rebuild_target_vectors()

    @property
    def n_tris(self):
        r"""
        The number of triangles in the triangle list.

        :type: int
        """
        return self.source.n_tris

    @property
    def trilist(self):
        r"""
        The triangle list.

        :type: (`n_tris`, 3) ndarray
        """
        return self.source.trilist

    def _rebuild_target_vectors(self):
        r"""
        Rebuild the vectors that are used in the apply method. This needs to
        be called whenever the target is changed.
        """
        t = self.target.points[self.trilist]
        # get vectors ij ik for the target
        self.tij, self.tik = t[:, 1] - t[:, 0], t[:, 2] - t[:, 0]
        # target i'th vertex positions
        self.ti = t[:, 0]

    def _sync_state_from_target(self):
        r"""
        PWA is particularly efficient to sync from target - we don't have to
        do much at all, just rebuild the target vectors.
        """
        self._rebuild_target_vectors()

    def _apply(self, x, **kwargs):
        """
        Applies this transform to a new set of vectors.

        Parameters
        ----------
        x : (K, 2) ndarray
            Points to apply this transform to.

        Returns
        -------
        transformed : (K, 2) ndarray
            The transformed array.
        """
        tri_index, alpha, beta = self.index_alpha_beta(x)
        return (self.ti[tri_index] +
                alpha[:, None] * self.tij[tri_index] +
                beta[:, None] * self.tik[tri_index])

    @abc.abstractmethod
    def index_alpha_beta(self, points):
        """
        Finds for each input point the index of it's bounding triangle
        and the alpha and beta value for that point in the triangle. Note
        this means that the following statements will always be true:
            alpha + beta <= 1
            alpha >= 0
            beta >= 0
        for each triangle result.
        Trying to map a point that does not exist in a
        triangle throws a TriangleContainmentError.

        Parameters
        ----------
        points : (K, 2) ndarray
            Points to test.

        Returns
        -------
        tri_index : (L,) ndarray
            triangle index for each of the `points`, assigning each
            point to it's containing triangle.
        alpha : (L,) ndarray
            Alpha for containing triangle of each point.
        beta : (L,) ndarray
            Beta for containing triangle of each point.

        Raises
        ------
        TriangleContainmentError
        All `points` must be contained in a source triangle. Check
        `error.points_outside_source_domain` to handle this case.
        """
        pass

    @property
    def has_true_inverse(self):
        return True

    def pseudoinverse(self):
        from menpo.shape import PointCloud, TriMesh  # to avoid circular import
        new_source = TriMesh(self.target.points, self.source.trilist)
        new_target = PointCloud(self.source.points)
        return type(self)(new_source, new_target)


class PythonPWA(AbstractPWA):

    def __init__(self, source, target):
        super(PythonPWA, self).__init__(source, target)
        si, sij, sik = barycentric_vectors(self.source.points, self.trilist)
        self.s, self.sij, self.sik = si, sij, sik

    def index_alpha_beta(self, points):
        return index_alpha_beta(self.s, self.sij, self.sik, points)


class CachedPWA(PythonPWA):

    def __init__(self, source, target):
        super(CachedPWA, self).__init__(source, target)
        self._applied_points, self._iab = None, None

    def index_alpha_beta(self, points):
        if (self._applied_points is None or not
                np.all(points == self._applied_points)):
            self._applied_points = points
            self._iab = PythonPWA.index_alpha_beta(self, points)
        return self._iab


class CythonPWA(AbstractPWA):
    r"""
    A piecewise affine transformation.

    The apply method in this case involves dotting the triangle vectors with
    the values of alpha and beta found. The calculation of alpha and beta is
     done in C, and a hash map is used to cache lookup values.

    Parameters
    ----------
    source : :class:`menpo.shape.PointCloud` or :class:`menpo.shape.TriMesh`
        The source points. If a TriMesh is provided, the triangulation on
        the TriMesh is used. If a :class:`menpo.shape.PointCloud`
        is provided, a Delaunay triangulation of the source is performed
        automatically.
    target : :class:`PointCloud`
        The target points. Note that the trilist is entirely decided by
        the source.

    Raises
    ------
    ValueError
        Source and target must both be 2D.

    TriangleContainmentError
        All points to apply must be contained in a source triangle. Check
        `error.points_outside_source_domain` to handle this case.
    """
    def __init__(self, source, target):
        super(CythonPWA, self).__init__(source, target)
        # make sure the source and target satisfy the c requirements
        source_c = np.require(self.source.points, dtype=np.float64,
                              requirements=['C'])
        trilist_c = np.require(self.trilist, dtype=np.uint32,
                               requirements=['C'])
        # build the cython wrapped C object and store it locally
        self._fastpwa = CLookupPWA(source_c, trilist_c)

    def copy(self):
        new = Copyable.copy(self)
        new._fastpwa = deepcopy(self._fastpwa)
        return new

    def index_alpha_beta(self, points):
        points_c = np.require(points, dtype=np.float64, requirements=['C'])
        index, alpha, beta = self._fastpwa.index_alpha_beta(points_c)
        if np.any(index < 0):
            raise TriangleContainmentError(index < 0)
        else:
            return index, alpha, beta



from nose.tools import assert_equals

import menpo.io as mio
from menpo.landmark import labeller, ibug_face_68
from menpo.shape import PointCloud


def test_squared_even_patches():
    image = mio.import_builtin_asset('breakingbad.jpg')
    patch_shape = (16, 16)
    patches = image.extract_patches(image.landmarks['PTS'].lms,
                                    patch_size=patch_shape)
    assert_equals(len(patches), 68)


def test_squared_odd_patches():
    image = mio.import_builtin_asset('breakingbad.jpg')
    patch_shape = (15, 15)
    patches = image.extract_patches(image.landmarks['PTS'].lms,
                                    patch_size=patch_shape)
    assert_equals(len(patches), 68)


def test_nonsquared_even_patches():
    image = mio.import_builtin_asset('breakingbad.jpg')
    patch_shape = (16, 18)
    patches = image.extract_patches(image.landmarks['PTS'].lms,
                                    patch_size=patch_shape)
    assert_equals(len(patches), 68)


def test_nonsquared_odd_patches():
    image = mio.import_builtin_asset('breakingbad.jpg')
    patch_shape = (15, 17)
    patches = image.extract_patches(image.landmarks['PTS'].lms,
                                    patch_size=patch_shape)
    assert_equals(len(patches), 68)


def test_nonsquared_even_odd_patches():
    image = mio.import_builtin_asset('breakingbad.jpg')
    patch_shape = (15, 16)
    patches = image.extract_patches(image.landmarks['PTS'].lms,
                                    patch_size=patch_shape)
    assert_equals(len(patches), 68)


def test_squared_even_patches_landmarks():
    image = mio.import_builtin_asset('breakingbad.jpg')
    patch_shape = (16, 16)
    patches = image.extract_patches_around_landmarks('PTS',
                                                     patch_size=patch_shape)
    assert_equals(len(patches), 68)


def test_squared_even_patches_landmarks_label():
    image = mio.import_builtin_asset('breakingbad.jpg')
    image = labeller(image, 'PTS', ibug_face_68)
    patch_shape = (16, 16)
    patches = image.extract_patches_around_landmarks('ibug_face_68',
                                                     label='nose',
                                                     patch_size=patch_shape)
    assert_equals(len(patches), 9)


def test_squared_even_patches_single_array():
    image = mio.import_builtin_asset('breakingbad.jpg')
    image = labeller(image, 'PTS', ibug_face_68)
    patch_shape = (16, 16)
    patches = image.extract_patches(image.landmarks['PTS'].lms,
                                    as_single_array=True,
                                    patch_size=patch_shape)
    assert_equals(patches.shape, ((68,) + patch_shape + (3,)))


def test_squared_even_patches_sample_offsets():
    image = mio.import_builtin_asset('breakingbad.jpg')
    image = labeller(image, 'PTS', ibug_face_68)
    sample_offsets = PointCloud([[0, 0], [1, 0]])
    patches = image.extract_patches(image.landmarks['PTS'].lms,
                                    sample_offsets=sample_offsets)
    assert_equals(len(patches), 136)



import numpy as np
from collections import Sized

from menpo.visualize.viewmatplotlib import (MatplotlibImageViewer2d,
                                            sample_colours_from_colourmap)

from .options import (channel_options, format_channel_options,
                      update_channel_options,
                      landmark_options, format_landmark_options,
                      update_landmark_options, info_print, format_info_print,
                      animation_options, format_animation_options,
                      save_figure_options, format_save_figure_options,
                      features_options, format_features_options, viewer_options,
                      format_viewer_options)
from .tools import logo, format_logo

# This glyph import is called frequently during visualisation, so we ensure
# that we only import it once
glyph = None


def visualize_pointclouds(pointclouds, figure_size=(10, 8), popup=False,
                          browser_style='buttons'):
    r"""
    Widget that allows browsing through a `list` of :map:`PointCloud`,
    :map:`PointGraph` or :map:`TriMesh` or subclasses.

    The widget has options tabs regarding the renderer (lines, markers, figure,
    axes) and saving the figure to file.

    Parameters
    ----------
    pointclouds : `list` of :map:`PointCloud` or :map:`PointGraph` or :map:`TriMesh` or subclasses
        The `list` of objects to be visualized.
    figure_size : (`int`, `int`), optional
        The initial size of the rendered figure.
    popup : `bool`, optional
        If ``True``, the widget will appear as a popup window.
    browser_style : {``buttons``, ``slider``}, optional
        It defines whether the selector of the objects will have the form of
        plus/minus buttons or a slider.
    """
    import IPython.display as ipydisplay
    import IPython.html.widgets as ipywidgets
    print('Initializing...')

    # make sure that pointclouds is a list even with one pointcloud member
    if not isinstance(pointclouds, Sized):
        pointclouds = [pointclouds]

    # find number of pointclouds
    n_pointclouds = len(pointclouds)

    # initial options dictionaries
    lines_options = {'render_lines': True,
                     'line_width': 1,
                     'line_colour': ['r'],
                     'line_style': '-'}
    markers_options = {'render_markers': True,
                       'marker_size': 20,
                       'marker_face_colour': ['r'],
                       'marker_edge_colour': ['k'],
                       'marker_style': 'o',
                       'marker_edge_width': 1}
    figure_options = {'x_scale': 1.,
                      'y_scale': 1.,
                      'render_axes': False,
                      'axes_font_name': 'sans-serif',
                      'axes_font_size': 10,
                      'axes_font_style': 'normal',
                      'axes_font_weight': 'normal',
                      'axes_x_limits': None,
                      'axes_y_limits': None}
    viewer_options_default = {'lines': lines_options,
                              'markers': markers_options,
                              'figure': figure_options}
    index_selection_default = {'min': 0,
                               'max': n_pointclouds-1,
                               'step': 1,
                               'index': 0}

    # define plot function
    def plot_function(name, value):
        import matplotlib.pyplot as plt
        # clear current figure, but wait until the new data to be displayed are
        # generated
        ipydisplay.clear_output(wait=True)

        # get selected pointcloud number
        im = 0
        if n_pointclouds > 1:
            im = pointcloud_number_wid.selected_values['index']

        # update info text widget
        update_info(pointclouds[im])

        # show pointcloud with selected options
        tmp1 = viewer_options_wid.selected_values[0]['lines']
        tmp2 = viewer_options_wid.selected_values[0]['markers']
        tmp3 = viewer_options_wid.selected_values[0]['figure']
        new_figure_size = (tmp3['x_scale'] * figure_size[0],
                           tmp3['y_scale'] * figure_size[1])

        renderer = pointclouds[im].view(
            figure_id=save_figure_wid.renderer[0].figure_id,
            new_figure=False, image_view=axes_mode_wid.value == 1,
            render_lines=tmp1['render_lines'],
            line_colour=tmp1['line_colour'][0],
            line_style=tmp1['line_style'], line_width=tmp1['line_width'],
            render_markers=tmp2['render_markers'],
            marker_style=tmp2['marker_style'], marker_size=tmp2['marker_size'],
            marker_face_colour=tmp2['marker_face_colour'][0],
            marker_edge_colour=tmp2['marker_edge_colour'][0],
            marker_edge_width=tmp2['marker_edge_width'],
            render_axes=tmp3['render_axes'],
            axes_font_name=tmp3['axes_font_name'],
            axes_font_size=tmp3['axes_font_size'],
            axes_font_style=tmp3['axes_font_style'],
            axes_font_weight=tmp3['axes_font_weight'],
            axes_x_limits=tmp3['axes_x_limits'],
            axes_y_limits=tmp3['axes_y_limits'],
            figure_size=new_figure_size,
            label=None)

        plt.show()

        # save the current figure id
        save_figure_wid.renderer[0] = renderer

    # define function that updates info text
    def update_info(pointcloud):
        min_b, max_b = pointcloud.bounds()
        rang = pointcloud.range()
        cm = pointcloud.centre()
        info_wid.children[1].children[0].value = "> {} points.".\
            format(pointcloud.n_points)
        info_wid.children[1].children[1].value = "> Bounds: " \
                                                 "[{0:.1f}-{1:.1f}]W, " \
                                                 "[{2:.1f}-{3:.1f}]H.".\
            format(min_b[0], max_b[0], min_b[1], max_b[1])
        info_wid.children[1].children[2].value = "> Range: {0:.1f}W, " \
                                                 "{1:.1f}H.".\
            format(rang[0], rang[1])
        info_wid.children[1].children[3].value = "> Centre of mass: " \
                                                 "({0:.1f}, {1:.1f}).".\
            format(cm[0], cm[1])

    # viewer options widget
    axes_mode_wid = ipywidgets.RadioButtonsWidget(
        values={'Image': 1, 'Point cloud': 2}, description='Axes mode:',
        value=2)
    axes_mode_wid.on_trait_change(plot_function, 'value')
    viewer_options_wid = viewer_options(viewer_options_default,
                                        ['lines', 'markers', 'figure_one'],
                                        objects_names=None,
                                        plot_function=plot_function,
                                        toggle_show_visible=False,
                                        toggle_show_default=True)
    viewer_options_all = ipywidgets.ContainerWidget(children=[axes_mode_wid,
                                                    viewer_options_wid])
    info_wid = info_print(n_bullets=4, toggle_show_default=True,
                          toggle_show_visible=False)

    # save figure widget
    initial_renderer = MatplotlibImageViewer2d(figure_id=None, new_figure=True,
                                               image=np.zeros((10, 10)))
    save_figure_wid = save_figure_options(initial_renderer,
                                          toggle_show_default=True,
                                          toggle_show_visible=False)

    # create final widget
    if n_pointclouds > 1:
        # pointcloud selection slider
        pointcloud_number_wid = animation_options(
            index_selection_default, plot_function=plot_function,
            index_description='Pointcloud Number', index_minus_description='<',
            index_plus_description='>', index_style=browser_style,
            index_text_editable=True, loop_default=True, interval_default=0.3,
            toggle_show_title='Pointcloud Options', toggle_show_default=True,
            toggle_show_visible=False)

        # final widget
        logo_wid = ipywidgets.ContainerWidget(children=[logo(),
                                                        pointcloud_number_wid])
        button_title = 'Pointclouds Menu'
    else:
        # final widget
        logo_wid = logo()
        button_title = 'Pointcloud Menu'
    # create popup widget if asked
    cont_wid = ipywidgets.TabWidget(children=[info_wid, viewer_options_all,
                                              save_figure_wid])
    if popup:
        wid = ipywidgets.PopupWidget(children=[logo_wid, cont_wid],
                                     button_text=button_title)
    else:
        wid = ipywidgets.ContainerWidget(children=[logo_wid, cont_wid])

    # display final widget
    ipydisplay.display(wid)

    # set final tab titles
    tab_titles = ['Info', 'Viewer options', 'Save figure']
    for (k, tl) in enumerate(tab_titles):
        wid.children[1].set_title(k, tl)

    # align-start the pointcloud number widget and the rest
    if n_pointclouds > 1:
        wid.add_class('align-start')

    # format options' widgets
    if n_pointclouds > 1:
        wid.children[0].remove_class('vbox')
        wid.children[0].add_class('hbox')
        format_animation_options(pointcloud_number_wid,
                                 index_text_width='1.0cm',
                                 container_padding='6px',
                                 container_margin='6px',
                                 container_border='1px solid black',
                                 toggle_button_font_weight='bold',
                                 border_visible=False)
    format_viewer_options(viewer_options_wid, container_padding='6px',
                          container_margin='6px',
                          container_border='1px solid black',
                          toggle_button_font_weight='bold',
                          border_visible=False,
                          suboptions_border_visible=True)
    format_info_print(info_wid, font_size_in_pt='10pt', container_padding='6px',
                      container_margin='6px',
                      container_border='1px solid gray',
                      toggle_button_font_weight='bold', border_visible=False)
    format_save_figure_options(save_figure_wid, container_padding='6px',
                               container_margin='6px',
                               container_border='1px solid black',
                               toggle_button_font_weight='bold',
                               tab_top_margin='0cm', border_visible=False)

    # Reset value to trigger initial visualization
    axes_mode_wid.value = 1


def visualize_landmarkgroups(landmarkgroups, figure_size=(10, 8), popup=False,
                             browser_style='buttons'):
    r"""
    Widget that allows browsing through a `list` of :map:`LandmarkGroup`
    (or subclass) objects.

    The landmark groups can have a combination of different attributes, e.g.
    different labels, number of points etc. The widget has options tabs
    regarding the landmarks, the renderer (lines, markers, numbering, legend,
    figure, axes) and saving the figure to file.

    Parameters
    ----------
    landmarkgroups : `list` of :map:`LandmarkGroup` or subclass
        The `list` of landmark groups to be visualized.
    figure_size : (`int`, `int`), optional
        The initial size of the rendered figure.
    popup : `bool`, optional
        If ``True``, the widget will appear as a popup window.
    browser_style : {``buttons``, ``slider``}, optional
        It defines whether the selector of the landmark managers will have the
        form of plus/minus buttons or a slider.
    """
    import IPython.display as ipydisplay
    import IPython.html.widgets as ipywidgets
    print('Initializing...')

    # make sure that landmarkgroups is a list even with one landmark group
    # member
    if not isinstance(landmarkgroups, list):
        landmarkgroups = [landmarkgroups]

    # find number of landmarkgroups
    n_landmarkgroups = len(landmarkgroups)

    # find all available groups and the respective labels from all the landmarks
    # that are passed in
    all_groups = []
    all_labels = []
    groups_l = 0
    for i, l in enumerate(landmarkgroups):
        labels_l = l.labels
        if labels_l not in all_labels:
            groups_l += 1
            all_groups.append(str(groups_l))
            all_labels.append(labels_l)

    # get initial line colours for each available group
    line_colours = []
    for l in all_labels:
        if len(l) == 1:
            line_colours.append(['r'])
        else:
            line_colours.append(sample_colours_from_colourmap(len(l), 'jet'))

    # initial options dictionaries
    landmark_options_default = {'render_landmarks': True,
                                'group_keys': ['0'],
                                'labels_keys': [landmarkgroups[0].labels],
                                'group': None,
                                'with_labels': None}
    index_selection_default = {'min': 0,
                               'max': n_landmarkgroups-1,
                               'step': 1,
                               'index': 0}
    markers_options = {'render_markers': True,
                       'marker_size': 20,
                       'marker_face_colour': ['r'],
                       'marker_edge_colour': ['k'],
                       'marker_style': 'o',
                       'marker_edge_width': 1}
    numbering_options = {'render_numbering': False,
                         'numbers_font_name': 'sans-serif',
                         'numbers_font_size': 10,
                         'numbers_font_style': 'normal',
                         'numbers_font_weight': 'normal',
                         'numbers_font_colour': ['k'],
                         'numbers_horizontal_align': 'center',
                         'numbers_vertical_align': 'bottom'}
    legend_options = {'render_legend': True,
                      'legend_title': '',
                      'legend_font_name': 'sans-serif',
                      'legend_font_style': 'normal',
                      'legend_font_size': 10,
                      'legend_font_weight': 'normal',
                      'legend_marker_scale': 1.,
                      'legend_location': 2,
                      'legend_bbox_to_anchor': (1.05, 1.),
                      'legend_border_axes_pad': 1.,
                      'legend_n_columns': 1,
                      'legend_horizontal_spacing': 1.,
                      'legend_vertical_spacing': 1.,
                      'legend_border': True,
                      'legend_border_padding': 0.5,
                      'legend_shadow': False,
                      'legend_rounded_corners': False}
    figure_options = {'x_scale': 1.,
                      'y_scale': 1.,
                      'render_axes': False,
                      'axes_font_name': 'sans-serif',
                      'axes_font_size': 10,
                      'axes_font_style': 'normal',
                      'axes_font_weight': 'normal',
                      'axes_x_limits': None,
                      'axes_y_limits': None}
    viewer_options_default = []
    for i in range(len(all_labels)):
        lines_options_default = {'render_lines': True,
                                 'line_width': 1,
                                 'line_colour': line_colours[i],
                                 'line_style': '-'}
        tmp = {'lines': lines_options_default,
               'markers': markers_options,
               'numbering': numbering_options,
               'legend': legend_options,
               'figure': figure_options}
        viewer_options_default.append(tmp)

    # Define plot function
    def plot_function(name, value):
        import matplotlib.pyplot as plt
        # clear current figure, but wait until the new data to be displayed are
        # generated
        ipydisplay.clear_output(wait=True)

        # get selected pointcloud number
        im = 0
        if n_landmarkgroups > 1:
            im = landmark_number_wid.selected_values['index']

        # update info text widget
        update_info(landmarkgroups[im])

        # show landmarks with selected options
        group_idx = all_labels.index(landmarkgroups[im].labels)
        tmp1 = viewer_options_wid.selected_values[group_idx]['lines']
        tmp2 = viewer_options_wid.selected_values[group_idx]['markers']
        tmp3 = viewer_options_wid.selected_values[group_idx]['numbering']
        tmp4 = viewer_options_wid.selected_values[group_idx]['legend']
        tmp5 = viewer_options_wid.selected_values[group_idx]['figure']
        new_figure_size = (tmp5['x_scale'] * figure_size[0],
                           tmp5['y_scale'] * figure_size[1])
        n_labels = len(landmark_options_wid.selected_values['with_labels'])

        renderer = landmarkgroups[im].view(
            with_labels=landmark_options_wid.selected_values['with_labels'],
            figure_id=save_figure_wid.renderer[0].figure_id,
            new_figure=False, image_view=axes_mode_wid.value == 1,
            render_lines=tmp1['render_lines'],
            line_colour=tmp1['line_colour'][:n_labels],
            line_style=tmp1['line_style'], line_width=tmp1['line_width'],
            render_markers=tmp2['render_markers'],
            marker_style=tmp2['marker_style'], marker_size=tmp2['marker_size'],
            marker_face_colour=tmp2['marker_face_colour'][0],
            marker_edge_colour=tmp2['marker_edge_colour'][0],
            marker_edge_width=tmp2['marker_edge_width'],
            render_numbering=tmp3['render_numbering'],
            numbers_font_name=tmp3['numbers_font_name'],
            numbers_font_size=tmp3['numbers_font_size'],
            numbers_font_style=tmp3['numbers_font_style'],
            numbers_font_weight=tmp3['numbers_font_weight'],
            numbers_font_colour=tmp3['numbers_font_colour'][0],
            numbers_horizontal_align=tmp3['numbers_horizontal_align'],
            numbers_vertical_align=tmp3['numbers_vertical_align'],
            legend_n_columns=tmp4['legend_n_columns'],
            legend_border_axes_pad=tmp4['legend_border_axes_pad'],
            legend_rounded_corners=tmp4['legend_rounded_corners'],
            legend_title=tmp4['legend_title'],
            legend_horizontal_spacing=tmp4['legend_horizontal_spacing'],
            legend_shadow=tmp4['legend_shadow'],
            legend_location=tmp4['legend_location'],
            legend_font_name=tmp4['legend_font_name'],
            legend_bbox_to_anchor=tmp4['legend_bbox_to_anchor'],
            legend_border=tmp4['legend_border'],
            legend_marker_scale=tmp4['legend_marker_scale'],
            legend_vertical_spacing=tmp4['legend_vertical_spacing'],
            legend_font_weight=tmp4['legend_font_weight'],
            legend_font_size=tmp4['legend_font_size'],
            render_legend=tmp4['render_legend'],
            legend_font_style=tmp4['legend_font_style'],
            legend_border_padding=tmp4['legend_border_padding'],
            render_axes=tmp5['render_axes'],
            axes_font_name=tmp5['axes_font_name'],
            axes_font_size=tmp5['axes_font_size'],
            axes_font_style=tmp5['axes_font_style'],
            axes_font_weight=tmp5['axes_font_weight'],
            axes_x_limits=tmp5['axes_x_limits'],
            axes_y_limits=tmp5['axes_y_limits'],
            figure_size=new_figure_size)

        plt.show()

        # save the current figure id
        save_figure_wid.renderer[0] = renderer

    # define function that updates info text
    def update_info(landmarkgroup):
        min_b, max_b = landmarkgroup.values()[0].bounds()
        rang = landmarkgroup.values()[0].range()
        cm = landmarkgroup.values()[0].centre()

        info_wid.children[1].children[0].value = "> {} landmark points.".\
            format(landmarkgroup.values()[0].n_points)
        info_wid.children[1].children[1].value = "> Bounds: " \
                                                 "[{0:.1f}-{1:.1f}]W, " \
                                                 "[{2:.1f}-{3:.1f}]H.".\
            format(min_b[0], max_b[0], min_b[1], max_b[1])
        info_wid.children[1].children[2].value = "> Range: {0:.1f}W, " \
                                                 "{1:.1f}H.".\
            format(rang[0], rang[1])
        info_wid.children[1].children[3].value = "> Centre of mass: " \
                                                 "({0:.1f}, {1:.1f}).".\
            format(cm[0], cm[1])
        info_wid.children[1].children[4].value = "> Norm is {0:.2f}.".\
            format(landmarkgroup.values()[0].norm())

    # create options widgets
    # The landmarks checkbox default value if the first image doesn't have
    # landmarks
    landmark_options_wid = landmark_options(landmark_options_default,
                                            plot_function=plot_function,
                                            toggle_show_default=True,
                                            toggle_show_visible=False)

    # viewer options widget
    axes_mode_wid = ipywidgets.RadioButtonsWidget(
        values={'Image': 1, 'Point cloud': 2}, description='Axes mode:',
        value=2)
    axes_mode_wid.on_trait_change(plot_function, 'value')
    viewer_options_wid = viewer_options(viewer_options_default,
                                        ['lines', 'markers', 'numbering',
                                         'legend', 'figure_one'],
                                        objects_names=all_groups,
                                        plot_function=plot_function,
                                        toggle_show_visible=False,
                                        toggle_show_default=True,
                                        labels=all_labels)
    # make the selection dropdown invisible, as ti is controlled by the
    # landmarks selection
    viewer_options_wid.children[1].children[0].visible = False
    viewer_options_all = ipywidgets.ContainerWidget(children=[axes_mode_wid,
                                                    viewer_options_wid])
    info_wid = info_print(n_bullets=5,
                          toggle_show_default=True, toggle_show_visible=False)

    # save figure widget
    initial_renderer = MatplotlibImageViewer2d(figure_id=None, new_figure=True,
                                               image=np.zeros((10, 10)))
    save_figure_wid = save_figure_options(initial_renderer,
                                          toggle_show_default=True,
                                          toggle_show_visible=False)

    # define function that updates options' widgets state
    def update_widgets(name, value):
        # get new groups and labels, update landmark options and format them
        im = 0
        if n_landmarkgroups > 1:
            im = landmark_number_wid.selected_values['index']
        group_l = all_labels.index(landmarkgroups[im].labels)
        update_landmark_options(landmark_options_wid, [str(group_l)],
                                [landmarkgroups[im].labels], plot_function)
        format_landmark_options(landmark_options_wid, container_padding='6px',
                                container_margin='6px',
                                container_border='1px solid black',
                                toggle_button_font_weight='bold',
                                border_visible=False)

        # set correct group to viewer options' selection
        viewer_options_wid.children[1].children[0].value = \
            all_labels.index(landmarkgroups[im].labels)
    landmark_options_wid.children[2].children[0].on_trait_change(update_widgets,
                                                                 'value')

    # create final widget
    if n_landmarkgroups > 1:
        # landmark selection slider
        landmark_number_wid = animation_options(
            index_selection_default, plot_function=plot_function,
            update_function=update_widgets, index_description='Shape Number',
            index_minus_description='<', index_plus_description='>',
            index_style=browser_style, index_text_editable=True,
            loop_default=True, interval_default=0.3,
            toggle_show_title='Shape Options', toggle_show_default=True,
            toggle_show_visible=False)

        # final widget
        logo_wid = ipywidgets.ContainerWidget(children=[logo(),
                                                        landmark_number_wid])
        button_title = 'Shapes Menu'
    else:
        # final widget
        logo_wid = logo()
        button_title = 'Shape Menu'
    # create popup widget if asked
    cont_wid = ipywidgets.TabWidget(children=[info_wid, landmark_options_wid,
                                              viewer_options_all,
                                              save_figure_wid])
    if popup:
        wid = ipywidgets.PopupWidget(children=[logo_wid, cont_wid],
                                     button_text=button_title)
    else:
        wid = ipywidgets.ContainerWidget(children=[logo_wid, cont_wid])

    # display final widget
    ipydisplay.display(wid)

    # set final tab titles
    tab_titles = ['Info', 'Landmarks options', 'Viewer options',
                  'Save figure']
    for (k, tl) in enumerate(tab_titles):
        wid.children[1].set_title(k, tl)

    # align-start the image number widget and the rest
    if n_landmarkgroups > 1:
        wid.add_class('align-start')

    # format options' widgets
    if n_landmarkgroups > 1:
        wid.children[0].remove_class('vbox')
        wid.children[0].add_class('hbox')
        format_animation_options(landmark_number_wid, index_text_width='1.0cm',
                                 container_padding='6px',
                                 container_margin='6px',
                                 container_border='1px solid black',
                                 toggle_button_font_weight='bold',
                                 border_visible=False)
    format_landmark_options(landmark_options_wid, container_padding='6px',
                            container_margin='6px',
                            container_border='1px solid black',
                            toggle_button_font_weight='bold',
                            border_visible=False)
    format_viewer_options(viewer_options_wid, container_padding='6px',
                          container_margin='6px',
                          container_border='1px solid black',
                          toggle_button_font_weight='bold',
                          border_visible=False,
                          suboptions_border_visible=True)
    format_info_print(info_wid, font_size_in_pt='10pt', container_padding='6px',
                      container_margin='6px',
                      container_border='1px solid gray',
                      toggle_button_font_weight='bold', border_visible=False)
    format_save_figure_options(save_figure_wid, container_padding='6px',
                               container_margin='6px',
                               container_border='1px solid black',
                               toggle_button_font_weight='bold',
                               tab_top_margin='0cm', border_visible=False)

    # update widgets' state for shape number 0
    update_widgets('', 0)

    # Reset value to trigger initial visualization
    axes_mode_wid.value = 1


def visualize_landmarks(landmarks, figure_size=(10, 8), popup=False,
                        browser_style='buttons'):
    r"""
    Widget that allows browsing through a `list` of :map:`LandmarkManager`
    (or subclass) objects.

    The managers can have a combination of different attributes, e.g. different
    landmark groups and labels etc. The widget has options tabs regarding the
    landmarks, the renderer (lines, markers, numbering, legend, figure, axes)
    and saving the figure to file.

    Parameters
    ----------
    landmarks : `list` of :map:`LandmarkManager` or subclass
        The `list` of landmark managers to be visualized.
    figure_size : (`int`, `int`), optional
        The initial size of the rendered figure.
    popup : `bool`, optional
        If ``True``, the widget will appear as a popup window.
    browser_style : {``buttons``, ``slider``}, optional
        It defines whether the selector of the landmark managers will have the
        form of plus/minus buttons or a slider.
    """
    import IPython.display as ipydisplay
    import IPython.html.widgets as ipywidgets
    print('Initializing...')

    # make sure that landmarks (groups) is a list even with one landmark group
    # member
    if not isinstance(landmarks, list):
        landmarks = [landmarks]

    # find number of landmarks
    n_landmarks = len(landmarks)

    # find initial groups and labels that will be passed to the landmark options
    # widget creation
    first_has_landmarks = landmarks[0].n_groups != 0
    if first_has_landmarks:
        initial_groups_keys, initial_labels_keys = \
            _extract_group_labels_landmarks(landmarks[0])
    else:
        initial_groups_keys = [' ']
        initial_labels_keys = [[' ']]

    # find all available groups and the respective labels from all the landmarks
    # that are passed in
    all_groups = []
    all_labels = []
    for l in landmarks:
        groups_l, labels_l = _extract_group_labels_landmarks(l)
        for i, g in enumerate(groups_l):
            if g not in all_groups:
                all_groups.append(g)
                all_labels.append(labels_l[i])

    # get initial line colours for each available group
    line_colours = []
    for l in all_labels:
        if len(l) == 1:
            line_colours.append(['r'])
        else:
            line_colours.append(sample_colours_from_colourmap(len(l), 'jet'))

    # initial options dictionaries
    landmark_options_default = {'render_landmarks': first_has_landmarks,
                                'group_keys': initial_groups_keys,
                                'labels_keys': initial_labels_keys,
                                'group': None,
                                'with_labels': None}
    index_selection_default = {'min': 0,
                               'max': n_landmarks-1,
                               'step': 1,
                               'index': 0}
    markers_options = {'render_markers': True,
                       'marker_size': 20,
                       'marker_face_colour': ['r'],
                       'marker_edge_colour': ['k'],
                       'marker_style': 'o',
                       'marker_edge_width': 1}
    numbering_options = {'render_numbering': False,
                         'numbers_font_name': 'sans-serif',
                         'numbers_font_size': 10,
                         'numbers_font_style': 'normal',
                         'numbers_font_weight': 'normal',
                         'numbers_font_colour': ['k'],
                         'numbers_horizontal_align': 'center',
                         'numbers_vertical_align': 'bottom'}
    legend_options = {'render_legend': True,
                      'legend_title': '',
                      'legend_font_name': 'sans-serif',
                      'legend_font_style': 'normal',
                      'legend_font_size': 10,
                      'legend_font_weight': 'normal',
                      'legend_marker_scale': 1.,
                      'legend_location': 2,
                      'legend_bbox_to_anchor': (1.05, 1.),
                      'legend_border_axes_pad': 1.,
                      'legend_n_columns': 1,
                      'legend_horizontal_spacing': 1.,
                      'legend_vertical_spacing': 1.,
                      'legend_border': True,
                      'legend_border_padding': 0.5,
                      'legend_shadow': False,
                      'legend_rounded_corners': False}
    figure_options = {'x_scale': 1.,
                      'y_scale': 1.,
                      'render_axes': False,
                      'axes_font_name': 'sans-serif',
                      'axes_font_size': 10,
                      'axes_font_style': 'normal',
                      'axes_font_weight': 'normal',
                      'axes_x_limits': None,
                      'axes_y_limits': None}
    viewer_options_default = []
    for i in range(len(all_groups)):
        lines_options_default = {'render_lines': True,
                                 'line_width': 1,
                                 'line_colour': line_colours[i],
                                 'line_style': '-'}
        tmp = {'lines': lines_options_default,
               'markers': markers_options,
               'numbering': numbering_options,
               'legend': legend_options,
               'figure': figure_options}
        viewer_options_default.append(tmp)

    # Define plot function
    def plot_function(name, value):
        import matplotlib.pyplot as plt
        # clear current figure, but wait until the new data to be displayed are
        # generated
        ipydisplay.clear_output(wait=True)

        # get selected pointcloud number
        im = 0
        if n_landmarks > 1:
            im = landmark_number_wid.selected_values['index']

        # update info text widget
        update_info(landmarks[im],
                    landmark_options_wid.selected_values['group'])

        # show landmarks with selected options
        group_idx = all_groups.index(landmark_options_wid.selected_values['group'])
        tmp1 = viewer_options_wid.selected_values[group_idx]['lines']
        tmp2 = viewer_options_wid.selected_values[group_idx]['markers']
        tmp3 = viewer_options_wid.selected_values[group_idx]['numbering']
        tmp4 = viewer_options_wid.selected_values[group_idx]['legend']
        tmp5 = viewer_options_wid.selected_values[group_idx]['figure']
        new_figure_size = (tmp5['x_scale'] * figure_size[0],
                           tmp5['y_scale'] * figure_size[1])
        n_labels = len(landmark_options_wid.selected_values['with_labels'])

        sel_group = landmark_options_wid.selected_values['group']
        renderer = landmarks[im][sel_group].view(
            with_labels=landmark_options_wid.selected_values['with_labels'],
            figure_id=save_figure_wid.renderer[0].figure_id,
            new_figure=False, image_view=axes_mode_wid.value == 1,
            render_lines=tmp1['render_lines'],
            line_colour=tmp1['line_colour'][:n_labels],
            line_style=tmp1['line_style'], line_width=tmp1['line_width'],
            render_markers=tmp2['render_markers'],
            marker_style=tmp2['marker_style'], marker_size=tmp2['marker_size'],
            marker_face_colour=tmp2['marker_face_colour'][0],
            marker_edge_colour=tmp2['marker_edge_colour'][0],
            marker_edge_width=tmp2['marker_edge_width'],
            render_numbering=tmp3['render_numbering'],
            numbers_font_name=tmp3['numbers_font_name'],
            numbers_font_size=tmp3['numbers_font_size'],
            numbers_font_style=tmp3['numbers_font_style'],
            numbers_font_weight=tmp3['numbers_font_weight'],
            numbers_font_colour=tmp3['numbers_font_colour'][0],
            numbers_horizontal_align=tmp3['numbers_horizontal_align'],
            numbers_vertical_align=tmp3['numbers_vertical_align'],
            legend_n_columns=tmp4['legend_n_columns'],
            legend_border_axes_pad=tmp4['legend_border_axes_pad'],
            legend_rounded_corners=tmp4['legend_rounded_corners'],
            legend_title=tmp4['legend_title'],
            legend_horizontal_spacing=tmp4['legend_horizontal_spacing'],
            legend_shadow=tmp4['legend_shadow'],
            legend_location=tmp4['legend_location'],
            legend_font_name=tmp4['legend_font_name'],
            legend_bbox_to_anchor=tmp4['legend_bbox_to_anchor'],
            legend_border=tmp4['legend_border'],
            legend_marker_scale=tmp4['legend_marker_scale'],
            legend_vertical_spacing=tmp4['legend_vertical_spacing'],
            legend_font_weight=tmp4['legend_font_weight'],
            legend_font_size=tmp4['legend_font_size'],
            render_legend=tmp4['render_legend'],
            legend_font_style=tmp4['legend_font_style'],
            legend_border_padding=tmp4['legend_border_padding'],
            render_axes=tmp5['render_axes'],
            axes_font_name=tmp5['axes_font_name'],
            axes_font_size=tmp5['axes_font_size'],
            axes_font_style=tmp5['axes_font_style'],
            axes_font_weight=tmp5['axes_font_weight'],
            axes_x_limits=tmp5['axes_x_limits'],
            axes_y_limits=tmp5['axes_y_limits'],
            figure_size=new_figure_size)

        plt.show()

        # save the current figure id
        save_figure_wid.renderer[0] = renderer

    # define function that updates info text
    def update_info(landmarks, group):
        if group != ' ':
            min_b, max_b = landmarks[group][None].bounds()
            rang = landmarks[group][None].range()
            cm = landmarks[group][None].centre()
            info_wid.children[1].children[0].value = "> {} landmark points.".\
                format(landmarks[group][None].n_points)
            info_wid.children[1].children[1].value = "> Bounds: " \
                                                     "[{0:.1f}-{1:.1f}]W, " \
                                                     "[{2:.1f}-{3:.1f}]H.".\
                format(min_b[0], max_b[0], min_b[1], max_b[1])
            info_wid.children[1].children[2].value = "> Range: {0:.1f}W, " \
                                                     "{1:.1f}H.".\
                format(rang[0], rang[1])
            info_wid.children[1].children[3].value = "> Centre of mass: " \
                                                     "({0:.1f}, {1:.1f}).".\
                format(cm[0], cm[1])
            info_wid.children[1].children[4].value = "> Norm is {0:.2f}.".\
                format(landmarks[group][None].norm())
        else:
            info_wid.children[1].children[0].value = "There are no landmarks."
            info_wid.children[1].children[1].value = ""
            info_wid.children[1].children[2].value = ""
            info_wid.children[1].children[3].value = ""
            info_wid.children[1].children[4].value = ""

    # create options widgets
    # The landmarks checkbox default value if the first image doesn't have
    # landmarks
    landmark_options_wid = landmark_options(landmark_options_default,
                                            plot_function=plot_function,
                                            toggle_show_default=True,
                                            toggle_show_visible=False)
    # if only a single landmark manager is passed in and it doesn't have
    # landmarks, then landmarks checkbox should be disabled
    landmark_options_wid.children[1].disabled = not first_has_landmarks

    # viewer options widget
    axes_mode_wid = ipywidgets.RadioButtonsWidget(
        values={'Image': 1, 'Point cloud': 2}, description='Axes mode:',
        value=2)
    axes_mode_wid.on_trait_change(plot_function, 'value')
    viewer_options_wid = viewer_options(viewer_options_default,
                                        ['lines', 'markers', 'numbering',
                                         'legend', 'figure_one'],
                                        objects_names=all_groups,
                                        plot_function=plot_function,
                                        toggle_show_visible=False,
                                        toggle_show_default=True,
                                        labels=all_labels)
    # make the selection dropdown invisible, as ti is controlled by the
    # landmarks selection
    viewer_options_wid.children[1].children[0].visible = False
    viewer_options_all = ipywidgets.ContainerWidget(children=[axes_mode_wid,
                                                    viewer_options_wid])
    info_wid = info_print(n_bullets=5,
                          toggle_show_default=True, toggle_show_visible=False)

    # save figure widget
    initial_renderer = MatplotlibImageViewer2d(figure_id=None, new_figure=True,
                                               image=np.zeros((10, 10)))
    save_figure_wid = save_figure_options(initial_renderer,
                                          toggle_show_default=True,
                                          toggle_show_visible=False)

    # define function that updates options' widgets state
    def update_widgets(name, value):
        # get new groups and labels, update landmark options and format them
        im = 0
        if n_landmarks > 1:
            im = landmark_number_wid.selected_values['index']
        group_keys, labels_keys = _extract_group_labels_landmarks(landmarks[im])
        update_landmark_options(landmark_options_wid, group_keys,
                                labels_keys, plot_function)
        format_landmark_options(landmark_options_wid, container_padding='6px',
                                container_margin='6px',
                                container_border='1px solid black',
                                toggle_button_font_weight='bold',
                                border_visible=False)

        # change group on viewer_options
        viewer_options_wid.children[1].children[0].value = \
            all_groups.index(landmark_options_wid.selected_values['group'])

    # create final widget
    if n_landmarks > 1:
        # landmark selection slider
        landmark_number_wid = animation_options(
            index_selection_default, plot_function=plot_function,
            update_function=update_widgets, index_description='Shape Number',
            index_minus_description='<', index_plus_description='>',
            index_style=browser_style, index_text_editable=True,
            loop_default=True, interval_default=0.3,
            toggle_show_title='Shape Options', toggle_show_default=True,
            toggle_show_visible=False)

        # final widget
        logo_wid = ipywidgets.ContainerWidget(children=[logo(),
                                                        landmark_number_wid])
        button_title = 'Shapes Menu'
    else:
        # final widget
        logo_wid = logo()
        button_title = 'Shape Menu'
    # create popup widget if asked
    cont_wid = ipywidgets.TabWidget(children=[info_wid, landmark_options_wid,
                                              viewer_options_all,
                                              save_figure_wid])
    if popup:
        wid = ipywidgets.PopupWidget(children=[logo_wid, cont_wid],
                                     button_text=button_title)
    else:
        wid = ipywidgets.ContainerWidget(children=[logo_wid, cont_wid])

    # display final widget
    ipydisplay.display(wid)

    # set final tab titles
    tab_titles = ['Info', 'Landmarks options', 'Viewer options',
                  'Save figure']
    for (k, tl) in enumerate(tab_titles):
        wid.children[1].set_title(k, tl)

    # align-start the image number widget and the rest
    if n_landmarks > 1:
        wid.add_class('align-start')

    # update viewer options
    def update_viewer_options(name, value):
        # set correct group at viewer options' selection
        if cont_wid.selected_index == 2:
            viewer_options_wid.children[1].children[0].value = \
                all_groups.index(landmark_options_wid.selected_values['group'])
    cont_wid.on_trait_change(update_viewer_options, 'selected_index')

    # format options' widgets
    if n_landmarks > 1:
        wid.children[0].remove_class('vbox')
        wid.children[0].add_class('hbox')
        format_animation_options(landmark_number_wid, index_text_width='1.0cm',
                                 container_padding='6px',
                                 container_margin='6px',
                                 container_border='1px solid black',
                                 toggle_button_font_weight='bold',
                                 border_visible=False)
    format_landmark_options(landmark_options_wid, container_padding='6px',
                            container_margin='6px',
                            container_border='1px solid black',
                            toggle_button_font_weight='bold',
                            border_visible=False)
    format_viewer_options(viewer_options_wid, container_padding='6px',
                          container_margin='6px',
                          container_border='1px solid black',
                          toggle_button_font_weight='bold',
                          border_visible=False,
                          suboptions_border_visible=True)
    format_info_print(info_wid, font_size_in_pt='10pt', container_padding='6px',
                      container_margin='6px',
                      container_border='1px solid gray',
                      toggle_button_font_weight='bold', border_visible=False)
    format_save_figure_options(save_figure_wid, container_padding='6px',
                               container_margin='6px',
                               container_border='1px solid black',
                               toggle_button_font_weight='bold',
                               tab_top_margin='0cm', border_visible=False)

    # update widgets' state for image number 0
    update_widgets('', 0)

    # Reset value to trigger initial visualization
    axes_mode_wid.value = 1


def visualize_images(images, figure_size=(10, 8), popup=False,
                     browser_style='buttons'):
    r"""
    Widget that allows browsing through a `list` of :map:`Image` (or subclass)
    objects.

    The images can have a combination of different attributes, e.g. masked or
    not, landmarked or not, without multiple landmark groups and labels etc.
    The widget has options tabs regarding the visualized channels, the
    landmarks, the renderer (lines, markers, numbering, legend, figure, axes)
    and saving the figure to file.

    Parameters
    ----------
    images : `list` of :map:`Image` or subclass
        The `list` of images to be visualized.
    figure_size : (`int`, `int`), optional
        The initial size of the rendered figure.
    popup : `bool`, optional
        If ``True``, the widget will appear as a popup window.
    browser_style : {``buttons``, ``slider``}, optional
        It defines whether the selector of the images will have the form of
        plus/minus buttons or a slider.
    """
    from menpo.image import MaskedImage
    import IPython.display as ipydisplay
    import IPython.html.widgets as ipywidgets
    print('Initializing...')

    # make sure that images is a list even with one image member
    if not isinstance(images, Sized):
        images = [images]

    # find number of images
    n_images = len(images)

    # find initial groups and labels that will be passed to the landmark options
    # widget creation
    first_has_landmarks = images[0].landmarks.n_groups != 0
    if first_has_landmarks:
        initial_groups_keys, initial_labels_keys = \
            _extract_groups_labels(images[0])
    else:
        initial_groups_keys = [' ']
        initial_labels_keys = [[' ']]

    # find all available groups and the respective labels from all the landmarks
    # that are passed in
    all_groups = []
    all_labels = []
    for l in images:
        groups_l, labels_l = _extract_groups_labels(l)
        for i, g in enumerate(groups_l):
            if g not in all_groups:
                all_groups.append(g)
                all_labels.append(labels_l[i])

    # get initial line colours for each available group
    line_colours = []
    for l in all_labels:
        if len(l) == 1:
            line_colours.append(['r'])
        else:
            line_colours.append(sample_colours_from_colourmap(len(l), 'jet'))

    # initial options dictionaries
    channels_default = 0
    if images[0].n_channels == 3:
        channels_default = None
    index_selection_default = {'min': 0,
                               'max': n_images-1,
                               'step': 1,
                               'index': 0}
    channels_options_default = {'n_channels': images[0].n_channels,
                                'image_is_masked': isinstance(images[0],
                                                              MaskedImage),
                                'channels': channels_default,
                                'glyph_enabled': False,
                                'glyph_block_size': 3,
                                'glyph_use_negative': False,
                                'sum_enabled': False,
                                'masked_enabled': isinstance(images[0],
                                                             MaskedImage)}
    landmark_options_default = {'render_landmarks': first_has_landmarks,
                                'group_keys': initial_groups_keys,
                                'labels_keys': initial_labels_keys,
                                'group': None,
                                'with_labels': None}
    image_options = {'interpolation': 'bilinear',
                     'alpha': 1.0}
    markers_options = {'render_markers': True,
                       'marker_size': 20,
                       'marker_face_colour': ['r'],
                       'marker_edge_colour': ['k'],
                       'marker_style': 'o',
                       'marker_edge_width': 1}
    numbering_options = {'render_numbering': False,
                         'numbers_font_name': 'sans-serif',
                         'numbers_font_size': 10,
                         'numbers_font_style': 'normal',
                         'numbers_font_weight': 'normal',
                         'numbers_font_colour': ['k'],
                         'numbers_horizontal_align': 'center',
                         'numbers_vertical_align': 'bottom'}
    legend_options = {'render_legend': True,
                      'legend_title': '',
                      'legend_font_name': 'sans-serif',
                      'legend_font_style': 'normal',
                      'legend_font_size': 10,
                      'legend_font_weight': 'normal',
                      'legend_marker_scale': 1.,
                      'legend_location': 2,
                      'legend_bbox_to_anchor': (1.05, 1.),
                      'legend_border_axes_pad': 1.,
                      'legend_n_columns': 1,
                      'legend_horizontal_spacing': 1.,
                      'legend_vertical_spacing': 1.,
                      'legend_border': True,
                      'legend_border_padding': 0.5,
                      'legend_shadow': False,
                      'legend_rounded_corners': False}
    figure_options = {'x_scale': 1.,
                      'y_scale': 1.,
                      'render_axes': False,
                      'axes_font_name': 'sans-serif',
                      'axes_font_size': 10,
                      'axes_font_style': 'normal',
                      'axes_font_weight': 'normal',
                      'axes_x_limits': None,
                      'axes_y_limits': None}
    viewer_options_default = []
    for i in range(len(all_groups)):
        lines_options_default = {'render_lines': True,
                                 'line_width': 1,
                                 'line_colour': line_colours[i],
                                 'line_style': '-'}
        tmp = {'lines': lines_options_default,
               'markers': markers_options,
               'numbering': numbering_options,
               'legend': legend_options,
               'figure': figure_options,
               'image': image_options}
        viewer_options_default.append(tmp)

    # define plot function
    def plot_function(name, value):
        # clear current figure, but wait until the new data to be displayed are
        # generated
        ipydisplay.clear_output(wait=True)

        # get selected image number
        im = 0
        if n_images > 1:
            im = image_number_wid.selected_values['index']

        # update info text widget
        image_has_landmarks = images[im].landmarks.n_groups != 0
        image_is_masked = isinstance(images[im], MaskedImage)
        update_info(images[im], image_is_masked, image_has_landmarks,
                    landmark_options_wid.selected_values['group'])
        n_labels = len(landmark_options_wid.selected_values['with_labels'])

        # show image with selected options
        group_idx = all_groups.index(landmark_options_wid.selected_values['group'])
        tmp1 = viewer_options_wid.selected_values[group_idx]['lines']
        tmp2 = viewer_options_wid.selected_values[group_idx]['markers']
        tmp3 = viewer_options_wid.selected_values[group_idx]['numbering']
        tmp4 = viewer_options_wid.selected_values[group_idx]['legend']
        tmp5 = viewer_options_wid.selected_values[group_idx]['figure']
        tmp6 = viewer_options_wid.selected_values[group_idx]['image']
        new_figure_size = (tmp5['x_scale'] * figure_size[0],
                           tmp5['y_scale'] * figure_size[1])
        renderer = _visualize(
            images[im], save_figure_wid.renderer[0],
            landmark_options_wid.selected_values['render_landmarks'],
            channel_options_wid.selected_values['image_is_masked'],
            channel_options_wid.selected_values['masked_enabled'],
            channel_options_wid.selected_values['channels'],
            channel_options_wid.selected_values['glyph_enabled'],
            channel_options_wid.selected_values['glyph_block_size'],
            channel_options_wid.selected_values['glyph_use_negative'],
            channel_options_wid.selected_values['sum_enabled'],
            landmark_options_wid.selected_values['group'],
            landmark_options_wid.selected_values['with_labels'],
            tmp1['render_lines'], tmp1['line_style'], tmp1['line_width'],
            tmp1['line_colour'][:n_labels], tmp2['render_markers'],
            tmp2['marker_style'], tmp2['marker_size'],
            tmp2['marker_edge_width'], tmp2['marker_edge_colour'][0],
            tmp2['marker_face_colour'][0], tmp3['render_numbering'],
            tmp3['numbers_font_name'], tmp3['numbers_font_size'],
            tmp3['numbers_font_style'], tmp3['numbers_font_weight'],
            tmp3['numbers_font_colour'][0], tmp3['numbers_horizontal_align'],
            tmp3['numbers_vertical_align'], tmp4['legend_n_columns'],
            tmp4['legend_border_axes_pad'], tmp4['legend_rounded_corners'],
            tmp4['legend_title'], tmp4['legend_horizontal_spacing'],
            tmp4['legend_shadow'], tmp4['legend_location'],
            tmp4['legend_font_name'], tmp4['legend_bbox_to_anchor'],
            tmp4['legend_border'], tmp4['legend_marker_scale'],
            tmp4['legend_vertical_spacing'], tmp4['legend_font_weight'],
            tmp4['legend_font_size'], tmp4['render_legend'],
            tmp4['legend_font_style'], tmp4['legend_border_padding'],
            new_figure_size, tmp5['render_axes'], tmp5['axes_font_name'],
            tmp5['axes_font_size'], tmp5['axes_font_style'],
            tmp5['axes_x_limits'], tmp5['axes_y_limits'],
            tmp5['axes_font_weight'], tmp6['interpolation'], tmp6['alpha'])

        # save the current figure id
        save_figure_wid.renderer[0] = renderer

    # define function that updates info text
    def update_info(img, image_is_masked, image_has_landmarks, group):
        # Prepare masked (or non-masked) string
        masked_str = 'Masked Image' if image_is_masked else 'Image'
        # get image path, if available
        path_str = img.path if hasattr(img, 'path') else 'No path available.'
        # Display masked pixels if image is masked
        masked_pixels_str = (r'{} masked pixels (attached mask {:.1%} true)'.
                             format(img.n_true_pixels(),
                                    img.mask.proportion_true())
                             if image_is_masked else '')
        # Display number of landmarks if image is landmarked
        landmarks_str = (r'{} landmark points.'.
                         format(img.landmarks[group].lms.n_points)
                         if image_has_landmarks else '')

        info_wid.children[1].children[0].value = "> {} of size {} with {} " \
                                                 "channel{}".\
            format(masked_str, img._str_shape, img.n_channels,
                   's' * (img.n_channels > 1))
        info_wid.children[1].children[1].value = "> Path: '{}'".format(path_str)
        info_wid.children[1].children[2].visible = image_is_masked
        info_wid.children[1].children[2].value = "> {}".format(masked_pixels_str)
        info_wid.children[1].children[3].value = "> min={:.3f}, max={:.3f}.".\
            format(img.pixels.min(), img.pixels.max())
        info_wid.children[1].children[4].visible = image_has_landmarks
        info_wid.children[1].children[4].value = "> {}".format(landmarks_str)

    # channel options widget
    channel_options_wid = channel_options(channels_options_default,
                                          plot_function=plot_function,
                                          toggle_show_default=True,
                                          toggle_show_visible=False)

    # landmarks options widget
    # The landmarks checkbox default value if the first image doesn't have
    # landmarks
    landmark_options_wid = landmark_options(landmark_options_default,
                                            plot_function=plot_function,
                                            toggle_show_default=True,
                                            toggle_show_visible=False)
    # if only a single image is passed in and it doesn't have landmarks, then
    # landmarks checkbox should be disabled
    landmark_options_wid.children[1].disabled = not first_has_landmarks

    # viewer options widget
    viewer_options_wid = viewer_options(viewer_options_default,
                                        ['lines', 'markers', 'numbering',
                                         'legend', 'figure_one', 'image'],
                                        objects_names=all_groups,
                                        plot_function=plot_function,
                                        toggle_show_visible=False,
                                        toggle_show_default=True,
                                        labels=all_labels)
    # make the selection dropdown invisible, as ti is controlled by the
    # landmarks selection
    viewer_options_wid.children[1].children[0].visible = False
    info_wid = info_print(n_bullets=5, toggle_show_default=True,
                          toggle_show_visible=False)

    # save figure widget
    initial_renderer = MatplotlibImageViewer2d(figure_id=None, new_figure=True,
                                               image=np.zeros((10, 10)))
    save_figure_wid = save_figure_options(initial_renderer,
                                          toggle_show_default=True,
                                          toggle_show_visible=False)

    # define function that updates options' widgets state
    def update_widgets(name, value):
        # set channels = 0 to make sure that when plotting is triggered by the
        # update_landmark_options, we don't get an error for not enough channels
        tmp_channels = channel_options_wid.selected_values['channels']
        channel_options_wid.selected_values['channels'] = 0

        # get new groups and labels, update landmark options and format them
        im = 0
        if n_images > 1:
            im = image_number_wid.selected_values['index']
        group_keys, labels_keys = _extract_groups_labels(images[im])
        update_landmark_options(landmark_options_wid, group_keys,
                                labels_keys, plot_function)
        format_landmark_options(landmark_options_wid,
                                container_padding='6px',
                                container_margin='6px',
                                container_border='1px solid black',
                                toggle_button_font_weight='bold',
                                border_visible=False)

        # update channel options
        channel_options_wid.selected_values['channels'] = tmp_channels
        update_channel_options(channel_options_wid,
                               n_channels=images[im].n_channels,
                               image_is_masked=isinstance(images[im],
                                                          MaskedImage))

        # set correct group at viewer options' selection
        if not group_keys == [' ']:
            viewer_options_wid.children[1].children[0].value = \
                all_groups.index(landmark_options_wid.selected_values['group'])

    # create final widget
    if n_images > 1:
        # image selection slider
        image_number_wid = animation_options(index_selection_default,
                                             plot_function=plot_function,
                                             update_function=update_widgets,
                                             index_description='Image Number',
                                             index_minus_description='<',
                                             index_plus_description='>',
                                             index_style=browser_style,
                                             index_text_editable=True,
                                             loop_default=True,
                                             interval_default=0.3,
                                             toggle_show_title='Image Options',
                                             toggle_show_default=True,
                                             toggle_show_visible=False)

        # final widget
        logo_wid = ipywidgets.ContainerWidget(children=[logo(),
                                                        image_number_wid])
        button_title = 'Images Menu'
    else:
        # final widget
        logo_wid = logo()
        button_title = 'Image Menu'
    # create popup widget if asked
    cont_wid = ipywidgets.TabWidget(children=[info_wid, channel_options_wid,
                                              landmark_options_wid,
                                              viewer_options_wid,
                                              save_figure_wid])
    if popup:
        wid = ipywidgets.PopupWidget(children=[logo_wid, cont_wid],
                                     button_text=button_title)
    else:
        wid = ipywidgets.ContainerWidget(children=[logo_wid, cont_wid])

    # display final widget
    ipydisplay.display(wid)

    # set final tab titles
    tab_titles = ['Info', 'Channels options', 'Landmarks options',
                  'Viewer options', 'Save figure']
    for (k, tl) in enumerate(tab_titles):
        wid.children[1].set_title(k, tl)

    # update viewer options
    def update_viewer_options(name, value):
        # set correct group at viewer options' selection
        if cont_wid.selected_index == 3:
            viewer_options_wid.children[1].children[0].value = \
                all_groups.index(landmark_options_wid.selected_values['group'])
    cont_wid.on_trait_change(update_viewer_options, 'selected_index')

    # align-start the image number widget and the rest
    if n_images > 1:
        wid.add_class('align-start')

    # format options' widgets
    if n_images > 1:
        wid.children[0].remove_class('vbox')
        wid.children[0].add_class('hbox')
        format_animation_options(image_number_wid, index_text_width='1.0cm',
                                 container_padding='6px',
                                 container_margin='6px',
                                 container_border='1px solid black',
                                 toggle_button_font_weight='bold',
                                 border_visible=False)
    format_channel_options(channel_options_wid, container_padding='6px',
                           container_margin='6px',
                           container_border='1px solid black',
                           toggle_button_font_weight='bold',
                           border_visible=False)
    format_landmark_options(landmark_options_wid, container_padding='6px',
                            container_margin='6px',
                            container_border='1px solid black',
                            toggle_button_font_weight='bold',
                            border_visible=False)
    format_viewer_options(viewer_options_wid, container_padding='6px',
                          container_margin='6px',
                          container_border='1px solid black',
                          toggle_button_font_weight='bold',
                          border_visible=False,
                          suboptions_border_visible=True)
    format_info_print(info_wid, font_size_in_pt='10pt', container_padding='6px',
                      container_margin='6px',
                      container_border='1px solid gray',
                      toggle_button_font_weight='bold', border_visible=False)
    format_save_figure_options(save_figure_wid, container_padding='6px',
                               container_margin='6px',
                               container_border='1px solid black',
                               toggle_button_font_weight='bold',
                               tab_top_margin='0cm', border_visible=False)

    # update widgets' state for image number 0
    update_widgets('', 0)

    # Reset value to trigger initial visualization
    viewer_options_wid.children[1].children[1].children[3].children[1].children[0].value = False


def save_matplotlib_figure(renderer, popup=True):
    r"""
    Widget that allows to save a figure, which was generated with Matplotlib,
    to file.

    Parameters
    ----------
    renderer : :map:`MatplotlibRenderer`
        The Matplotlib renderer object.
    popup : `bool`, optional
        If ``True``, the widget will appear as a popup window.
    """
    import IPython.display as ipydisplay
    import IPython.html.widgets as ipywidgets
    # Create sub-widgets
    logo_wid = logo()
    save_figure_wid = save_figure_options(renderer, toggle_show_default=True,
                                          toggle_show_visible=False)

    # Create final widget
    if popup:
        wid = ipywidgets.PopupWidget(children=[logo_wid, save_figure_wid],
                                     button_text='Save Figure')
        # set width of popup widget
        wid.set_css({'width': '11cm'}, selector='modal')
    else:
        wid = ipywidgets.ContainerWidget(children=[logo_wid, save_figure_wid])

    # Display widget
    ipydisplay.display(wid)

    # Format widgets
    format_save_figure_options(save_figure_wid, container_padding='6px',
                               container_margin='6px',
                               container_border='1px solid black',
                               toggle_button_font_weight='bold',
                               tab_top_margin='0cm', border_visible=True)
    format_logo(logo_wid, border_visible=False)


def features_selection(popup=True):
    r"""
    Widget that allows selecting a features function and its options. The
    widget supports all features from :ref:`api-feature-index` and has a
    preview tab. It returns a `list` of length 1 with the selected features
    function closure.

    Parameters
    ----------
    popup : `bool`, optional
        If ``True``, the widget will appear as a popup window.

    Returns
    -------
    features_function : `list` of length ``1``
        The function closure of the features function using `functools.partial`.
        So the function can be called as: ::

            features_image = features_function[0](image)

    """
    import IPython.display as ipydisplay
    import IPython.html.widgets as ipywidgets
    # Create sub-widgets
    logo_wid = logo()
    features_options_wid = features_options(toggle_show_default=True,
                                            toggle_show_visible=False)
    features_wid = ipywidgets.ContainerWidget(children=[logo_wid,
                                                        features_options_wid])
    select_but = ipywidgets.ButtonWidget(description='Select')

    # Create final widget
    if popup:
        wid = ipywidgets.PopupWidget(children=[features_wid, select_but],
                                     button_text='Features Selection')
    else:
        wid = ipywidgets.ContainerWidget(children=[features_wid, select_but])

    # function for select button
    def select_function(name):
        wid.close()
        output.pop(0)
        output.append(features_options_wid.function)
    select_but.on_click(select_function)

    # Display widget
    ipydisplay.display(wid)

    # Format widgets
    format_features_options(features_options_wid, border_visible=True)
    format_logo(logo_wid, border_visible=False)
    # set popup width
    if popup:
        wid.set_css({
            'width': '13cm'}, selector='modal')
    # align logo at the end
    features_wid.add_class('align-end')
    # align select button at the centre
    wid.add_class('align-center')

    # Initialize output with empty list. It needs to be a list so that
    # it's mutable and synchronizes with frontend.
    output = [features_options_wid.function]

    return output


def _visualize(image, renderer, render_landmarks, image_is_masked,
               masked_enabled, channels, glyph_enabled, glyph_block_size,
               glyph_use_negative, sum_enabled, group, with_labels,
               render_lines, line_style, line_width, line_colour,
               render_markers, marker_style, marker_size,
               marker_edge_width, marker_edge_colour, marker_face_colour,
               render_numbering, numbers_font_name, numbers_font_size,
               numbers_font_style, numbers_font_weight, numbers_font_colour,
               numbers_horizontal_align, numbers_vertical_align,
               legend_n_columns, legend_border_axes_pad, legend_rounded_corners,
               legend_title, legend_horizontal_spacing, legend_shadow,
               legend_location, legend_font_name, legend_bbox_to_anchor,
               legend_border, legend_marker_scale, legend_vertical_spacing,
               legend_font_weight, legend_font_size, render_legend,
               legend_font_style, legend_border_padding, figure_size,
               render_axes, axes_font_name, axes_font_size, axes_font_style,
               axes_x_limits, axes_y_limits, axes_font_weight, interpolation,
               alpha):
    import matplotlib.pyplot as plt

    global glyph
    if glyph is None:
        from menpo.visualize.image import glyph

    # This makes the code shorter for dealing with masked images vs non-masked
    # images
    mask_arguments = ({'masked': masked_enabled}
                      if image_is_masked else {})

    # plot
    if render_landmarks and not group == ' ':
        # show image with landmarks
        if glyph_enabled or sum_enabled:
            # image, landmarks, masked, glyph
            renderer = glyph(image, vectors_block_size=glyph_block_size,
                             use_negative=glyph_use_negative,
                             channels=channels).\
                view_landmarks(
                    group=group, with_labels=with_labels, without_labels=None,
                    figure_id=renderer.figure_id, new_figure=False,
                    render_lines=render_lines, line_colour=line_colour,
                    line_style=line_style, line_width=line_width,
                    render_markers=render_markers, marker_style=marker_style,
                    marker_size=marker_size,
                    marker_face_colour=marker_face_colour,
                    marker_edge_colour=marker_edge_colour,
                    marker_edge_width=marker_edge_width,
                    render_numbering=render_numbering,
                    numbers_horizontal_align=numbers_horizontal_align,
                    numbers_vertical_align=numbers_vertical_align,
                    numbers_font_name=numbers_font_name,
                    numbers_font_size=numbers_font_size,
                    numbers_font_style=numbers_font_style,
                    numbers_font_weight=numbers_font_weight,
                    numbers_font_colour=numbers_font_colour,
                    render_legend=render_legend, legend_title=legend_title,
                    legend_font_name=legend_font_name,
                    legend_font_style=legend_font_style,
                    legend_font_size=legend_font_size,
                    legend_font_weight=legend_font_weight,
                    legend_marker_scale=legend_marker_scale,
                    legend_location=legend_location,
                    legend_bbox_to_anchor=legend_bbox_to_anchor,
                    legend_border_axes_pad=legend_border_axes_pad,
                    legend_n_columns=legend_n_columns,
                    legend_horizontal_spacing=legend_horizontal_spacing,
                    legend_vertical_spacing=legend_vertical_spacing,
                    legend_border=legend_border,
                    legend_border_padding=legend_border_padding,
                    legend_shadow=legend_shadow,
                    legend_rounded_corners=legend_rounded_corners,
                    render_axes=render_axes, axes_font_name=axes_font_name,
                    axes_font_size=axes_font_size,
                    axes_font_style=axes_font_style,
                    axes_font_weight=axes_font_weight,
                    axes_x_limits=axes_x_limits, axes_y_limits=axes_y_limits,
                    figure_size=figure_size, interpolation=interpolation,
                    alpha=alpha, **mask_arguments)
        else:
            renderer = image.view_landmarks(
                channels=channels, group=group, with_labels=with_labels,
                without_labels=None, figure_id=renderer.figure_id,
                new_figure=False, render_lines=render_lines,
                line_colour=line_colour, line_style=line_style,
                line_width=line_width, render_markers=render_markers,
                marker_style=marker_style, marker_size=marker_size,
                marker_face_colour=marker_face_colour,
                marker_edge_colour=marker_edge_colour,
                marker_edge_width=marker_edge_width,
                render_numbering=render_numbering,
                numbers_horizontal_align=numbers_horizontal_align,
                numbers_vertical_align=numbers_vertical_align,
                numbers_font_name=numbers_font_name,
                numbers_font_size=numbers_font_size,
                numbers_font_style=numbers_font_style,
                numbers_font_weight=numbers_font_weight,
                numbers_font_colour=numbers_font_colour,
                render_legend=render_legend, legend_title=legend_title,
                legend_font_name=legend_font_name,
                legend_font_style=legend_font_style,
                legend_font_size=legend_font_size,
                legend_font_weight=legend_font_weight,
                legend_marker_scale=legend_marker_scale,
                legend_location=legend_location,
                legend_bbox_to_anchor=legend_bbox_to_anchor,
                legend_border_axes_pad=legend_border_axes_pad,
                legend_n_columns=legend_n_columns,
                legend_horizontal_spacing=legend_horizontal_spacing,
                legend_vertical_spacing=legend_vertical_spacing,
                legend_border=legend_border,
                legend_border_padding=legend_border_padding,
                legend_shadow=legend_shadow,
                legend_rounded_corners=legend_rounded_corners,
                render_axes=render_axes, axes_font_name=axes_font_name,
                axes_font_size=axes_font_size, axes_font_style=axes_font_style,
                axes_font_weight=axes_font_weight, axes_x_limits=axes_x_limits,
                axes_y_limits=axes_y_limits, figure_size=figure_size,
                interpolation=interpolation, alpha=alpha, **mask_arguments)
    else:
        # either there are not any landmark groups selected or they won't
        # be displayed
        if glyph_enabled or sum_enabled:
            # image, not landmarks, masked, glyph
            renderer = glyph(image, vectors_block_size=glyph_block_size,
                             use_negative=glyph_use_negative,
                             channels=channels).view(
                render_axes=render_axes, axes_font_name=axes_font_name,
                axes_font_size=axes_font_size, axes_font_style=axes_font_style,
                axes_font_weight=axes_font_weight, axes_x_limits=axes_x_limits,
                axes_y_limits=axes_y_limits, figure_size=figure_size,
                interpolation=interpolation, alpha=alpha, **mask_arguments)
        else:
            # image, not landmarks, masked, not glyph
            renderer = image.view(
                channels=channels, render_axes=render_axes,
                axes_font_name=axes_font_name, axes_font_size=axes_font_size,
                axes_font_style=axes_font_style,
                axes_font_weight=axes_font_weight, axes_x_limits=axes_x_limits,
                axes_y_limits=axes_y_limits, figure_size=figure_size,
                interpolation=interpolation, alpha=alpha, **mask_arguments)

    # show plot
    plt.show()

    return renderer


def _extract_groups_labels(image):
    r"""
    Function that extracts the groups and labels from an image's landmarks.

    Parameters
    ----------
    image : :map:`Image` or subclass
       The input image object.

    Returns
    -------
    group_keys : `list` of `str`
        The list of landmark groups found.

    labels_keys : `list` of `str`
        The list of lists of each landmark group's labels.
    """
    groups_keys, labels_keys = _extract_group_labels_landmarks(image.landmarks)
    return groups_keys, labels_keys


def _extract_group_labels_landmarks(landmark_manager):
    r"""
    Function that extracts the groups and labels from a landmark manager object.

    Parameters
    ----------
    landmark_manager : :map:`LandmarkManager` or subclass
       The input landmark manager object.

    Returns
    -------
    group_keys : `list` of `str`
        The list of landmark groups found.

    labels_keys : `list` of `str`
        The list of lists of each landmark group's labels.
    """
    groups_keys = landmark_manager.keys()
    if len(groups_keys) == 0:
        groups_keys = [' ']
        labels_keys = [[' ']]
    else:
        labels_keys = [landmark_manager[g].keys() for g in groups_keys]
    return groups_keys, labels_keys



import gzip
from functools import partial
from pathlib import Path

from .extensions import landmark_types, image_types, pickle_types
from ..utils import _norm_path

# an open file handle that uses a small fast level of compression
gzip_open = partial(gzip.open, compresslevel=3)


def export_landmark_file(landmark_group, fp, extension=None, overwrite=False):
    r"""
    Exports a given landmark group. The ``fp`` argument can be either
    or a `str` or any Python type that acts like a file. If a file is provided,
    the ``extension`` kwarg **must** be provided. If no
    ``extension`` is provided and a `str` filepath is provided, then
    the export type is calculated based on the filepath extension.

    Due to the mix in string and file types, an explicit overwrite argument is
    used which is ``False`` by default.

    Parameters
    ----------
    landmark_group : :map:`LandmarkGroup`
        The landmark group to export.
    fp : `str` or `file`-like object
        The string path or file-like object to save the object at/into.
    extension : `str` or None, optional
        The extension to use, this must match the file path if the file
        path is a string. Determines the type of exporter that is used.
    overwrite : `bool`, optional
        Whether or not to overwrite a file if it already exists.

    Raises
    ------
    ValueError
        File already exists and ``overwrite`` != ``True``
    ValueError
        ``fp`` is a `str` and the ``extension`` is not ``None``
        and the two extensions do not match
    ValueError
        ``fp`` is a `file`-like object and ``extension`` is
        ``None``
    ValueError
        The provided extension does not match to an existing exporter type
        (the output type is not supported).
    """
    _export(landmark_group, fp, landmark_types, extension, overwrite)


def export_image(image, fp, extension=None, overwrite=False):
    r"""
    Exports a given image. The ``fp`` argument can be either
    a `str` or any Python type that acts like a file. If a file is provided,
    the ``extension`` kwarg **must** be provided. If no
    ``extension`` is provided and a `str` filepath is provided, then
    the export type is calculated based on the filepath extension.

    Due to the mix of string and file types, an explicit overwrite argument is
    used which is ``False`` by default.

    Parameters
    ----------
    image : :map:`Image`
        The image to export.
    fp : `str` or `file`-like object
        The string path or file-like object to save the object at/into.
    extension : `str` or None, optional
        The extension to use, this must match the file path if the file
        path is a string. Determines the type of exporter that is used.
    overwrite : `bool`, optional
        Whether or not to overwrite a file if it already exists.

    Raises
    ------
    ValueError
        File already exists and ``overwrite`` != ``True``
    ValueError
        ``fp`` is a `str` and the ``extension`` is not ``None``
        and the two extensions do not match
    ValueError
        ``fp`` is a `file`-like object and ``extension`` is
        ``None``
    ValueError
        The provided extension does not match to an existing exporter type
        (the output type is not supported).
    """
    _export(image, fp, image_types, extension, overwrite)


def export_pickle(obj, fp, overwrite=False):
    r"""
    Exports a given collection of Python objects with Pickle.

    The ``fp`` argument can be either a `str` or any Python type that acts like
    a file.
    If ``fp`` is a path, it must have the suffix `.pkl` or `.pkl.gz`. If
    `.pkl`, the object will be pickled using Pickle protocol 2 without
    compression. If `.pkl.gz` the object will be pickled using Pickle protocol
    2 with gzip compression (at a fixed compression level of 3).

    Parameters
    ----------
    obj : ``object``
        The object to export.
    fp : `str` or `file`-like object
        The string path or file-like object to save the object at/into.
    overwrite : `bool`, optional
        Whether or not to overwrite a file if it already exists.

    Raises
    ------
    ValueError
        File already exists and ``overwrite`` != ``True``
    ValueError
        ``fp`` is a `file`-like object and ``extension`` is
        ``None``
    ValueError
        The provided extension does not match to an existing exporter type
        (the output type is not supported).
    """
    if isinstance(fp, Path):
        fp = str(fp)  # cheeky conversion to string to reuse existing code
    if isinstance(fp, basestring):
        # user provided a path - if it ended .gz we will compress
        path_filepath = _validate_filepath(fp, '.pkl', overwrite)
        o = gzip_open if path_filepath.suffix == '.gz' else open
        with o(fp, 'wb') as f:
            # force overwrite as True we've already done the check above
            _export(obj, f, pickle_types, '.pkl', True)
    else:
        _export(obj, fp, pickle_types, '.pkl', overwrite)


def _normalise_extension(extension):
    # Account for the fact the user may only have passed the extension
    # without the proceeding period
    if extension[0] is not '.':
        extension = '.' + extension
    return extension.lower()


def _extension_to_export_function(extension, extensions_map):
    try:
        extension = _normalise_extension(extension)
        return extensions_map[extension.lower()]
    except KeyError:
        raise ValueError('The output file extension provided is not currently '
                         'supported.')


def _validate_filepath(fp, extension, overwrite):
    path_filepath = Path(_norm_path(fp))
    if path_filepath.exists() and not overwrite:
        raise ValueError('File already exists. Please set the overwrite '
                         'kwarg if you wish to overwrite the file.')
    if extension is not None:
        # use .suffixes[0] to handle compression suffixes correctly (see below)
        filepath_suffix = path_filepath.suffixes[0]
        # we couldn't find an exporter for all the suffixes (e.g .foo.bar)
        # maybe the file stem has '.' in it? -> try again but this time just use the
        # final suffix (.bar). (Note we first try '.foo.bar' as we want to catch
        # cases like 'pkl.gz')
        if _normalise_extension(extension) != filepath_suffix and len(path_filepath.suffixes) > 1:
            filepath_suffix = path_filepath.suffix
        if _normalise_extension(extension) != filepath_suffix:
            raise ValueError('The file path extension must match the '
                             'requested file extension.')
    return path_filepath


def _export(obj, fp, extensions_map, extension, overwrite):
    if isinstance(fp, Path):
        fp = str(fp)  # cheeky conversion to string to reuse existing code
    if isinstance(fp, basestring):
        path_filepath = _validate_filepath(fp, extension, overwrite)

        export_function = _extension_to_export_function(
            path_filepath.suffix, extensions_map)

        with path_filepath.open('wb') as file_handle:
            export_function(obj, file_handle)
    else:
        # You MUST provide an extension if a file handle is given
        if extension is None:
            raise ValueError('An export file extension must be provided if a '
                             'file-like object is passed.')
        # Apparently in Python 2.x there is no reliable way to detect something
        # that is 'file' like (file handle or a StringIO object or something
        # you can read and write to like a file). Therefore, we are going to
        # just be really Pythonic about it and just assume we were handed
        # a correctly behaving object.
        try:
            # Follow PIL like behaviour. Check the file handle extension
            # and check if matches the given extension
            _validate_filepath(fp.name, extension, overwrite)
        except AttributeError:
            pass
        export_function = _extension_to_export_function(
            _normalise_extension(extension), extensions_map)
        export_function(obj, fp)
